# coding=UTF-8
# ex:ts=4:sw=4:et=on
#
# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import types
from functools import partial

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

from mvc.models.properties import (
    LabeledProperty, StringProperty, BoolProperty, FloatProperty,
    SignalMixin, ReadOnlyMixin, ListProperty
)
from mvc.observers import ListObserver
from mvc import Model

from pyxrd.generic.models import DataModel
from pyxrd.generic.io import storables, Storable, get_case_insensitive_glob

from pyxrd.refinement.refinables.mixins import RefinementValue
from pyxrd.refinement.refinables.properties import RefinableMixin
from pyxrd.refinement.refinables.metaclasses import PyXRDRefinableMeta

class ComponentPropMixin(object):
    """
        A mixin which provides some common utility functions for retrieving
        properties using a string description (e.g. 'layer_atoms.1' or 'b_cell')
    """

    def _parseattr(self, attr):
        """
            Function used for handling (deprecated) 'property strings':
            attr contains a string (e.g. cell_a or layer_atoms.2) which can be 
            parsed into an object and a property.
            Current implementation uses UUID's, however this is still here for
            backwards-compatibility...
            Will be removed at some point!
        """
        if not isinstance(attr, str):
            return attr

        if attr == "" or attr == None:
            return None

        def get_atom_by_index(atoms, index):
            for atom in atoms:
                if atom.atom_nr == index:
                    return atom
            return None

        attr = attr.replace("data_", "", 1) # for backwards compatibility
        attrs = attr.split(".")
        if attrs[0] == "layer_atoms":
            atom = get_atom_by_index(self.component._layer_atoms, int(attrs[1]))
            if atom is not None:
                return atom, "pn"
            else:
                return None
        elif attrs[0] == "interlayer_atoms":
            atom = get_atom_by_index(self.component._interlayer_atoms, int(attrs[1]))
            if atom is not None:
                return atom, "pn"
            else:
                return None
        else:
            return self.component, attr

@storables.register()
class AtomRelation(ComponentPropMixin, RefinementValue, DataModel, Storable, metaclass=PyXRDRefinableMeta):

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "AtomRelation"
        file_filters = [
            ("Atom relation", get_case_insensitive_glob("*.atr")),
        ]
        allowed_relations = {}

    component = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:
    #: The name of this AtomRelation
    name = StringProperty(
        default="", text="Name",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The value of this AtomRelation
    value = FloatProperty(
        default=0.0, text="Value",
        visible=True, persistent=True, tabular=True, widget_type='float_entry',
        signal_name="data_changed", refinable=True,
        mix_with=(SignalMixin, RefinableMixin)
    )

    #: Flag indicating whether this AtomRelation is enabled or not
    enabled = BoolProperty(
        default=True, text="Enabled",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: Is True when this AtomRelation's value is driven by another AtomRelation.
    #: Should never be set directly or things might break!
    driven_by_other = BoolProperty(
        default=False, text="Driven by other",
        visible=False, persistent=False, tabular=True
    )

    @property
    def applicable(self):
        """
        Is True when this AtomRelation was passed a component of which the atom
        ratios are not set to be inherited from another component.
        """
        return (self.parent is not None and not self.parent.inherit_atom_relations)

    # REFINEMENT VALUE IMPLEMENTATION:
    @property
    def refine_title(self):
        return self.name

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.component.phase.refine_title,
            component_name=self.component.refine_title
        )

    @property
    def refine_value(self):
        return self.value
    @refine_value.setter
    def refine_value(self, value):
        self.value = value

    @property
    def inside_linked_component(self):
        return (self.component.linked_with is not None) and self.component.inherit_atom_relations

    @property
    def is_refinable(self):
        return self.enabled and not self.driven_by_other and not self.inside_linked_component

    @property
    def refine_info(self):
        return self.value_ref_info

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Valid keyword arguments for an AtomRelation are:
                name: the name of this AtomRelation
                value: the value for this AtomRelation
                enabled: boolean indicating whether or not this AtomRelation is 
                 enabled
        """
        my_kwargs = self.pop_kwargs(kwargs,
            "data_name", "data_ratio", "ratio",
            *[prop.label for prop in AtomRelation.Meta.get_local_persistent_properties()]
        )
        super(AtomRelation, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        self.name = self.get_kwarg(kwargs, "", "name", "data_name")
        self.value = self.get_kwarg(kwargs, 0.0, "value", "ratio", "data_ratio")
        self.enabled = bool(self.get_kwarg(kwargs, True, "enabled"))

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def resolve_relations(self):
        raise NotImplementedError("Subclasses should implement the resolve_relations method!")

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def create_prop_store(self, prop=None):
        if self.component is not None:
            store = Gtk.ListStore(object, str, str)
            for atom in self.component._layer_atoms:
                store.append([atom, "pn", atom.name])
            for atom in self.component._interlayer_atoms:
                store.append([atom, "pn", atom.name])
            for relation in self.component._atom_relations:
                tp = relation.Meta.store_id
                if tp in self.Meta.allowed_relations:
                    for prop, name in self.Meta.allowed_relations[tp]:
                        if callable(name):
                            name = name(relation)
                        store.append([relation, prop, name])
            return store

    def iter_references(self):
        raise NotImplementedError("'iter_references' should be implemented by subclasses!")

    def _safe_is_referring(self, value):
        if value is not None and hasattr(value, "is_referring"):
            return value.is_referring([self, ])
        else:
            return False

    def is_referring(self, references=None):
        """
            Checks whether this AtomRelation is causing circular references.
            Can be used to check this before actually setting references by
            setting the 'references' keyword argument to a list containing the
            new reference value(s).
        """
        if references == None:
            references = []
        # 1. Bluntly check if we're not already somewhere referred to,
        #    if not, add ourselves to the list of references
        if self in references:
            return True
        references.append(self)

        # 2. Loop over our own references, check if they cause a circular
        #    reference, if not add them to the list of references.
        for reference in self.iter_references():
            if reference is not None and hasattr(reference, "is_referring"):
                if reference.is_referring(references):
                    return True
                else:
                    references.append(reference)

        return False

    def _set_driven_flag_for_prop(self, prop=None):
        """Internal method used to safely set the driven_by_other flag on an object.
        Subclasses can override to provide a check on the property set by the driver."""
        self.driven_by_other = True

    def apply_relation(self):
        raise NotImplementedError("Subclasses should implement the apply_relation method!")

    pass # end of class

@storables.register()
class AtomRatio(AtomRelation):

    # MODEL INTEL:
    class Meta(AtomRelation.Meta):
        store_id = "AtomRatio"
        allowed_relations = {
            "AtomRatio": [
                ("__internal_sum__", lambda o: "%s: SUM" % o.name),
                ("value", lambda o: "%s: RATIO" % o.name),
            ],
            "AtomContents": [("value", lambda o: o.name)],
        }

    # SIGNALS:

    # PROPERTIES:
    #: The sum of the two atoms
    sum = FloatProperty(
        default=1.0, text="Sum", minimum=0.0,
        visible=True, persistent=True, tabular=True, widget_type='float_entry',
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    def __internal_sum__(self, value):
        """
        Special setter for other AtomRelation objects depending on the value of
        the sum of the AtomRatio. This can be used to have multi-substitution by
        linking two (or more) AtomRatio's. Eg Al-by-Mg-&-Fe:
        
        AtomRatioMgAndFeForAl -> links together Al content and Fe+Mg content
        
                              => sum = e.g. 4 set by user
                              
        AtomRatioMgForFe      -> links together the Fe and Mg content
        
                              => sum = set by previous ratio.
        """
        self._sum = float(value)
        self.apply_relation()
    __internal_sum__ = property(fset=__internal_sum__)

    def _set_driven_flag_for_prop(self, prop, *args):
        """Internal method used to safely set the driven_by_other flag on an object.
        Subclasses can override to provide a check on the property set by the driver."""
        if prop != "__internal_sum__":
            super(AtomRatio, self)._set_driven_flag_for_prop(prop)

    def _set_atom(self, value, label=None):
        if not self._safe_is_referring(value[0]):
            getattr(type(self), label)._set(self, value)

    #: The substituting atom
    atom1 = LabeledProperty(
        default=[None, None], text="Substituting Atom",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", fset=partial(_set_atom, label="atom1"),
        mix_with=(SignalMixin,)
    )

    #: The Original atom
    atom2 = LabeledProperty(
        default=[None, None], text="Original Atom",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", fset=partial(_set_atom, label="atom2"),
        mix_with=(SignalMixin,)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs): # @ReservedAssignment
        """
            Valid keyword arguments for an AtomRatio are:
                sum: the sum of the atoms contents
                atom1: a tuple containing the first atom and its property name to read/set
                atom2: a tuple containing the first atom and its property name to read/set
            The value property is the 'ratio' of the first atom over the sum of both
        """
        my_kwargs = self.pop_kwargs(kwargs,
            "data_sum", "prop1", "data_prop1", "data_prop2", "prop2",
            *[prop.label for prop in AtomRatio.Meta.get_local_persistent_properties()]
        )
        super(AtomRatio, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        self.sum = self.get_kwarg(kwargs, self.sum, "sum", "data_sum")

        self._unresolved_atom1 = self._parseattr(self.get_kwarg(kwargs, [None, None], "atom1", "prop1", "data_prop1"))
        self._unresolved_atom2 = self._parseattr(self.get_kwarg(kwargs, [None, None], "atom2", "prop2", "data_prop2"))

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def json_properties(self):
        retval = Storable.json_properties(self)
        retval["atom1"] = [retval["atom1"][0].uuid if retval["atom1"][0] else None, retval["atom1"][1]]
        retval["atom2"] = [retval["atom2"][0].uuid if retval["atom2"][0] else None, retval["atom2"][1]]
        return retval

    def resolve_relations(self):
        if isinstance(self._unresolved_atom1[0], str):
            self._unresolved_atom1[0] = type(type(self)).object_pool.get_object(self._unresolved_atom1[0])
        self.atom1 = list(self._unresolved_atom1)
        del self._unresolved_atom1
        if isinstance(self._unresolved_atom2[0], str):
            self._unresolved_atom2[0] = type(type(self)).object_pool.get_object(self._unresolved_atom2[0])
        self.atom2 = list(self._unresolved_atom2)
        del self._unresolved_atom2

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def apply_relation(self):
        if self.enabled and self.applicable:
            for value, (atom, prop) in [(self.value, self.atom1), (1.0 - self.value, self.atom2)]:
                if atom and prop:
                    # do not fire events, just set attributes:
                    with atom.data_changed.ignore():
                        setattr(atom, prop, value * self.sum)
                        if hasattr(atom, "_set_driven_flag_for_prop"):
                            atom._set_driven_flag_for_prop(prop)

    def iter_references(self):
        for atom in [self.atom1[0], self.atom2[0]]:
            yield atom

    pass # end of class

class AtomContentObject(Model):
    """
        Wrapper around an atom object used in the AtomContents model.
        Stores the atom, the property to set and it's default amount.
    """

    atom = LabeledProperty(
        default=None, text="Atom",
        visible=False, persistent=False, tabular=True,
    )

    prop = LabeledProperty(
        default=None, text="Prop",
        visible=False, persistent=False, tabular=True,
    )

    amount = FloatProperty(
        default=0.0, text="Amount", minimum=0.0,
        visible=False, persistent=False, tabular=True,
    )

    def __init__(self, atom, prop, amount, *args, **kwargs):
        super(AtomContentObject, self).__init__(*args, **kwargs)
        self.atom = atom
        self.prop = prop
        self.amount = amount

    def update_atom(self, value):
        if not (self.atom == "" or self.atom is None or self.prop is None):
            with self.atom.data_changed.ignore():
                setattr(self.atom, self.prop, self.amount * value)
                if hasattr(self.atom, "_set_driven_flag_for_prop"):
                    self.atom._set_driven_flag_for_prop(self.prop)

    pass

@storables.register()
class AtomContents(AtomRelation):

    # MODEL INTEL:
    class Meta(AtomRelation.Meta):
        store_id = "AtomContents"
        allowed_relations = {
            "AtomRatio": [
                ("__internal_sum__", lambda o: "%s: SUM" % o.name),
                ("value", lambda o: "%s: RATIO" % o.name),
            ],
        }

    # SIGNALS:

    # PROPERTIES:
    atom_contents = ListProperty(
        default=None, text="Atom contents",
        visible=True, persistent=True, tabular=True,
        data_type=AtomContentObject,
        mix_with=(ReadOnlyMixin,)
    )

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Valid keyword arguments for an AtomContents are:
                atom_contents: a list of tuples containing the atom content 
                 object uuids, property names and default amounts 
        """
        my_kwargs = self.pop_kwargs(kwargs,
            *[prop.label for prop in AtomContents.Meta.get_local_persistent_properties()]
        )
        super(AtomContents, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        # Load atom contents:
        atom_contents = []
        for uuid, prop, amount in self.get_kwarg(kwargs, [], "atom_contents"):
            # uuid's are resolved when resolve_relations is called
            atom_contents.append(AtomContentObject(uuid, prop, amount))
        type(self).atom_contents._set(self, atom_contents)

        def on_change(*args):
            if self.enabled: # no need for updates otherwise
                self.data_changed.emit()

        self._atom_contents_observer = ListObserver(
            on_change, on_change,
            prop_name="atom_contents",
            model=self
        )

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def json_properties(self):
        retval = Storable.json_properties(self)
        retval["atom_contents"] = list([
            [
                atom_contents.atom.uuid if atom_contents.atom else None,
                atom_contents.prop,
                atom_contents.amount
            ] for atom_contents in retval["atom_contents"]
        ])
        return retval

    def resolve_relations(self):
        # Disable event dispatching to prevent infinite loops
        enabled = self.enabled
        self.enabled = False
        # Change rows with string references to objects (uuid's)
        for atom_content in self.atom_contents:
            if isinstance(atom_content.atom, str):
                atom_content.atom = type(type(self)).object_pool.get_object(atom_content.atom)
        # Set the flag to its original value
        self.enabled = enabled

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def apply_relation(self):
        if self.enabled and self.applicable:
            for atom_content in self.atom_contents:
                atom_content.update_atom(self.value)

    def set_atom_content_values(self, path, new_atom, new_prop):
        """    
            Convenience function that first checks if the new atom value will
            not cause a circular reference before actually setting it.
        """
        with self.data_changed.hold():
            atom_content = self.atom_contents[int(path[0])]
            if atom_content.atom != new_atom:
                old_atom = atom_content.atom
                atom_content.atom = None # clear...
                if not self._safe_is_referring(new_atom):
                    atom_content.atom = new_atom
                else:
                    atom_content.atom = old_atom
            else:
                atom_content.atom = None
            atom_content.prop = new_prop

    def iter_references(self):
        for atom_content in self.atom_contents:
            yield atom_content.atom

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

from matplotlib.figure import Figure

# TODO: move this to mvc somehow:
from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvasGTK

from pyxrd.generic.views import BaseView


class EditAtomTypeView(BaseView):
    builder = resource_filename(__name__, "glade/atoms.glade")
    top = "edit_atom_type"
    widget_format = "atom_%s"

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        BaseView.__init__(self, *args, **kwargs)

        self.graph_parent = self["view_graph"]
        self.setup_matplotlib_widget()

    def setup_matplotlib_widget(self):
        self.figure = Figure(dpi=72)

        self.plot = self.figure.add_subplot(111)
        self.figure.subplots_adjust(bottom=0.20)

        self.matlib_canvas = FigureCanvasGTK(self.figure)

        self.plot.autoscale_view()

        self.graph_parent.add(self.matlib_canvas)
        self.graph_parent.show_all()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update_figure(self, x, y):
        self.plot.cla()
        self.plot.plot(x, y, 'k-', aa=True)
        self.plot.set_ylabel('Scattering factor', size=14, weight="heavy")
        self.plot.set_xlabel('2θ', size=14, weight="heavy")
        self.plot.autoscale_view()
        if self.matlib_canvas is not None:
            self.matlib_canvas.draw()



import os

def get_case_insensitive_glob(*strings):
    '''Ex: '*.ora' => '*.[oO][rR][aA]' '''
    return ['*.%s' % ''.join(["[%s%s]" % (c.lower(), c.upper()) for c in string.split('.')[1]]) for string in strings]

def retrieve_lowercase_extension(glob):
    '''Ex: '*.[oO][rR][aA]' => '*.ora' '''
    return ''.join([ c.replace("[", "").replace("]", "")[:-1] for c in glob.split('][')])

# ======================================================================
# This is taken from python 2.6 (os.path.relpath is supported in 2.6)
#
# Copyright (c) 2001 Python Software Foundation; All Rights Reserved
# This code about relpath is covered by the Python Software Foundation
# (PSF) Agreement. See http://docs.python.org/license.html for details.
# ======================================================================
def __posix_relpath(path, start=os.curdir):
    """Return a relative version of a path"""

    if not path: raise ValueError("no path specified")

    start_list = os.path.abspath(start).split(os.sep)
    path_list = os.path.abspath(path).split(os.sep)

    # Work out how much of the filepath is shared by start and path.
    i = len(os.path.commonprefix([start_list, path_list]))

    rel_list = [os.pardir] * (len(start_list) - i) + path_list[i:]
    if not rel_list: return os.curdir
    return os.path.join(*rel_list)

# This is taken from python 2.6 (os.path.relpath is supported in 2.6)
# This is for windows
def __nt_relpath(path, start=os.curdir):
    """Return a relative version of a path"""

    if not path: raise ValueError("no path specified")

    start_list = os.path.abspath(start).split(os.sep)
    path_list = os.path.abspath(path).split(os.sep)
    if start_list[0].lower() != path_list[0].lower():
        unc_path, rest = os.path.splitunc(path)
        unc_start, rest = os.path.splitunc(start)
        if bool(unc_path) ^ bool(unc_start):
            raise ValueError("Cannot mix UNC and non-UNC paths (%s and %s)" \
                             % (path, start))
        else: raise ValueError("path is on drive %s, start on drive %s" \
                               % (path_list[0], start_list[0]))
    # Work out how much of the filepath is shared by start and path.
    for i in range(min(len(start_list), len(path_list))):
        if start_list[i].lower() != path_list[i].lower():
            break
        else: i += 1
        pass

    rel_list = [os.pardir] * (len(start_list) - i) + path_list[i:]
    if not rel_list: return os.curdir
    return os.path.join(*rel_list)
try:
    import os.path.relpath
    relpath = os.path.relpath
except ImportError:
    if os.name == 'nt':
        relpath = __nt_relpath
    else:
        relpath = __posix_relpath
        pass
    pass
# ======================================================================
# End of code covered by PSF License Agreement
# ======================================================================


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from functools import partial

import numpy as np
from matplotlib.ticker import FixedLocator, FixedFormatter

from pyxrd.calculations.goniometer import get_2t_from_nm, get_nm_from_2t
from pyxrd.data import settings

def _get_ticks():

    def create_flat_array(*arrays):
        arr = np.array([], dtype=float)
        for new_arr in arrays:
            arr = np.append(arr, new_arr)
        return arr

    minor_ticks_nm = create_flat_array(
        np.arange(6.0, 2.0, -0.5),
        np.arange(2.0, 1.0, -0.1),
        np.arange(1.0, 0.8, -0.05),
        np.arange(0.80, 0.00, -0.01)
    )
    major_ticks_nm = create_flat_array(
        np.arange(6.0, 1.0, -1.0),
        np.arange(2.0, 1.0, -0.5),
        np.arange(1.0, 0.8, -0.1),
        np.arange(0.80, 0.00, -0.05)
    )
    label_ticks_nm = create_flat_array(
        np.arange(6.0, 1.0, -2.0),
        np.arange(2.0, 1.0, -0.5),
        np.arange(1.0, 0.8, -0.1),
        np.arange(0.80, 0.40, -0.1),
        np.arange(0.40, 0.00, -0.05)
    )

    return minor_ticks_nm, major_ticks_nm, label_ticks_nm

def set_nm_ticks(axis, wavelength, xmin, xmax):
    """
        Sets the tick positions and labels for a nanometer x-axes using
        the given lower & upper limits and the wavelength
    """

    np_nm2a = np.vectorize(partial(get_2t_from_nm, wavelength=wavelength))

    def get_tick_labels(a, b):
        def in_close(value, arr):
            for val in arr:
                if np.isclose(value, val):
                    return True
            return False
        return [ "%g" % val if in_close(val, b) else "" for val in a ]

    minor_ticks_nm, major_ticks_nm, label_ticks_nm = _get_ticks()

    dmax = min(get_nm_from_2t(xmin, wavelength), 100) #limit this so we don't get an "infinite" scale
    dmin = get_nm_from_2t(xmax, wavelength)

    # Extract the part we need:
    selector = (minor_ticks_nm >= dmin) & (minor_ticks_nm <= dmax)
    minor_ticks_pos = np_nm2a(minor_ticks_nm[selector])
    selector = (major_ticks_nm >= dmin) & (major_ticks_nm <= dmax)
    major_ticks_pos = np_nm2a(major_ticks_nm[selector])

    major_ticks_labels = get_tick_labels(major_ticks_nm[selector], label_ticks_nm)

    # Set the ticks
    helper = axis.get_helper()

    helper.axis.minor.locator = FixedLocator(minor_ticks_pos)
    helper.axis.minor.formatter = FixedFormatter([""] * len(minor_ticks_pos))

    helper.axis.major.locator = FixedLocator(major_ticks_pos)
    helper.axis.major.formatter = FixedFormatter(major_ticks_labels)

    pass #end of func

def update_xaxis(axes, title='Angle (°2$\\theta$)', weight='heavy',
            rotation=0, ha="center", va="center",
            pad=0, size=16, nm_ticks=False, wavelength=None):

    axis = axes.axis["bottom"]

    axis.major_ticks.set_tick_out(True)
    axis.minor_ticks.set_tick_out(True)

    axis.label.set_text(title)
    axis.label.set_weight(weight)
    axis.label.set_size(size)
    axis.label.set_pad(pad)

    axis.major_ticklabels.set_visible(True)
    axis.major_ticklabels.set_rotation(rotation)
    axis.major_ticklabels.set_ha(ha)
    axis.major_ticklabels.set_va(va)

    if nm_ticks: set_nm_ticks(axis, wavelength, *axes.get_xlim())

def update_lim(axes, pos_setup, project):
    # Autoscale the view:
    axes.autoscale_view(tight=True)
    xmin, xmax = axes.get_xlim()

    axes.set_ylim(bottom=0, auto=True)

    # Adjust x limits if needed:
    if project is None or project.axes_xlimit == 0:
        xmin, xmax = max(xmin, 0.0), max(xmax, 20.0)
    else:
        xmin, xmax = max(project.axes_xmin, 0.0), project.axes_xmax
    axes.set_xlim(left=xmin, right=xmax, auto=False)


    # Adjust y limits if needed
    if project is not None and project.axes_ylimit != 0:
        scale, _ = project.get_scale_factor()

        ymin = max(project.axes_ymin, 0.0)
        ymax = project.axes_ymax
        if ymax <= 0:
            ymax = axes.get_ylim()[1]
        else:
            ymax = ymax * scale
        ymin = ymin * scale

        axes.set_ylim(bottom=ymin, top=ymax, auto=False)

    # Update plot position setup:
    pos_setup.xdiff = xmax - xmin
    pos_setup.xstretch = project.axes_xstretch if project is not None else False

def update_axes(axes, pos_setup, project, specimens):
    """
        Internal generic plot update method.
    """

    update_lim(axes, pos_setup, project)

    axes.axis["right"].set_visible(False)
    axes.axis["top"].set_visible(False)
    axes.get_xaxis().tick_bottom()
    axes.get_yaxis().tick_left()
    if project is None or not project.axes_yvisible:
        axes.axis["left"].set_visible(False)
    else:
        axes.axis["left"].set_visible(True)

    axes.set_position(pos_setup.position)

    if project is not None and project.axes_dspacing:
        if specimens is None or len(specimens) == 0:
            wavelength = settings.AXES_DEFAULT_WAVELENGTH
        else:
            wavelength = specimens[0].goniometer.wavelength
        update_xaxis(axes,
            title='d (nm)',
            rotation=-90, ha="left", pad=25,
            nm_ticks=True, wavelength=wavelength,
        )
    else:
        update_xaxis(axes)

class PositionSetup(object):
    """
        Keeps track of the positioning of a plot
    """
    left = settings.PLOT_LEFT
    top = settings.PLOT_TOP
    bottom = settings.PLOT_BOTTOM

    xdiff = 20
    xstretch = settings.AXES_XSTRETCH

    @property
    def right(self):
        return self.left + self.width

    @property
    def width(self):
        MAX_PLOT_WIDTH = settings.MAX_PLOT_RIGHT - self.left
        if self.xstretch:
            return MAX_PLOT_WIDTH
        else:
            return min((self.xdiff / 70), 1.0) * MAX_PLOT_WIDTH

    @property
    def height(self):
        return abs(self.top - self.bottom)

    @property
    def position(self):
        return [self.left, self.bottom, self.width, self.height]

    def to_string(self):
        return ":".join(map(str, [self.left,self.right,self.top,self.bottom,self.xdiff,self.xstretch])) 

    @property
    def default_bottom(self):
        return settings.PLOT_BOTTOM

    @property
    def default_left(self):
        return settings.PLOT_LEFT
    
    @property
    def default_top(self):
        return settings.PLOT_TOP

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2007 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import os
from uuid import uuid4 as get_uuid

def rec_setattr(obj, attr, value):
    """Set object's attribute. May use dot notation.

    >>> class C(object): pass
    >>> a = C()
    >>> a.b = C()
    >>> rec_setattr(a, 'b.c', 4)
    >>> a.b.c
    4
    """
    if obj is None:
        raise AttributeError("Cannot recursively set attribute (%s) on NoneType" % attr)
    else:
        if '.' not in attr:
            setattr(obj, attr, value)
        else:
            attr, attrs = attr.split('.', 1)
            rec_setattr(getattr(obj, attr), attrs, value)

def rec_getattr(obj, attr, default):
    """Get object's attribute. May use dot notation.

    >>> class C(object): pass
    >>> a = C()
    >>> a.b = C()
    >>> a.b.c = 4
    >>> rec_getattr(a, 'b.c')
    4
    """
    if obj is None:
        return default
    else:
        if '.' not in attr:
            return getattr(obj, attr, default)
        else:
            attr, attrs = attr.split('.', 1)
            return rec_getattr(getattr(obj, attr), attrs, default)

def round_sig(x, sig=1):
    if x == 0:
        return 0
    else:
        return round(x, sig - int(floor(log10(abs(x)))) - 1)

def not_none(passed, default):
    """Returns `passed` if not None, else `default` is returned"""
    return passed if passed is not None else default

def getmembers(_object, _predicate):
    """This is an implementation of inspect.getmembers, as in some versions 
    of python it may be buggy. 
    See issue at http://bugs.python.org/issue1785"""
    # This should be:
    # return inspect.getmembers(_object, _predicate)

    # ... and it is re-implemented as:
    observers = []
    for key in dir(_object):
        try: m = getattr(_object, key)
        except AttributeError: continue
        if _predicate(m): observers.append((key, m))
        pass
    return observers

def get_new_uuid():
    return str(get_uuid().hex)

def get_unique_list(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]

def pop_kwargs(kwargs, *keys):
    return {
        key: kwargs.pop(key) for key in keys if key in kwargs
    }


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk

def convert_string_to_gdk_color_int(color):
    """
        Converts a string hexadecimal color description to an integer that can
        be used by gdk functions.
    """
    color = Gdk.color_parse(color) # @UndefinedVariable
    return (int(color.red_float * 255) << 24) + (int(color.green_float * 255) << 16) + (int(color.blue_float * 255) << 8) + 255

def get_color_pb(color, width, height):
    """
        Gets a Gdk.Pixbuf filled with color (str)
    """
    pb = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, width, height)  # @UndefinedVariable
    pb.fill(convert_string_to_gdk_color_int(color))
    return pb


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.generic.controllers import BaseController

class EditInSituBehaviourController(BaseController):
    """
        The controller for the Behaviour model
    """

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

from .utils import adjust_filename_to_globs, retrieve_lowercase_extension, run_dialog
import os

class FileChooserDialog(Gtk.FileChooserDialog):

    accept_responses = (
        Gtk.ResponseType.ACCEPT, # @UndefinedVariable
        Gtk.ResponseType.YES, # @UndefinedVariable
        Gtk.ResponseType.APPLY, # @UndefinedVariable
        Gtk.ResponseType.OK # @UndefinedVariable
    )

    persist = False

    @property
    def parser(self):
        try:
            return getattr(self.get_filter(), "parser")
        except AttributeError:
            return None

    @property
    def filename(self):
        """ Extracts the selected filename from a Gtk.Dialog """
        filename = super(FileChooserDialog, self).get_filename()
        if filename is not None:
            filename = adjust_filename_to_globs(filename, self.selected_globs)
            self.set_filename(filename)
        return filename

    @property
    def selected_globs(self):
        """ Returns the extension glob corresponding to the selected filter """
        fltr = self.get_filter() # THIS RETURNS NONE??
        if fltr is None:
            return None
        else:
            selected_name = fltr.get_name()
            for fltr in self.filters:
                try:
                    name, globs = fltr
                except TypeError: # filter is not a tuple, perhaps it is a FileFilter from a parser
                    parser = getattr(fltr, "parser")
                    name, globs = parser.description, parser.extensions
                if selected_name == name:
                    if len(globs) and globs[0] != "*.*":
                        return [retrieve_lowercase_extension(glob) for glob in globs]
                    else:
                        return None

    def __init__(self, title, action, parent=None, buttons=None,
            current_name=None, current_folder=os.path.expanduser('~'),
            extra_widget=None, filters=[],
            multiple=False, confirm_overwrite=True, persist=False):
        super(FileChooserDialog, self).__init__(
            title=title, action=action, parent=parent, buttons=buttons
        )
        self.update(
            multiple=multiple, confirm_overwrite=confirm_overwrite,
            extra_widget=extra_widget, filters=filters,
            current_name=current_name, current_folder=current_folder,
            persist=persist
        )

    def update(self, **kwargs):
        """ Updates the dialog with the given set of keyword arguments, 
            and then returns itself """
        if "title" in kwargs and kwargs["title"] is not None:
            self.set_title(kwargs.pop("title"))

        if "action" in kwargs and kwargs["action"] is not None:
            self.set_action(kwargs.pop("action"))

        if "parent" in kwargs and kwargs["parent"] is not None:
            self.set_parent(kwargs.pop("parent"))

        if "buttons" in kwargs:
            self.get_action_area().foreach(lambda w: w.destroy())
            self.add_buttons(*kwargs.pop("buttons"))

        # Multiple files are allowed or not:
        if "multiple" in kwargs:
            self.set_select_multiple(kwargs.pop("multiple"))

        # Ask before overwriting yes/no
        if "confirm_overwrite" in kwargs:
            self.set_do_overwrite_confirmation(kwargs.pop("confirm_overwrite"))

        # Extra widget packed at the bottom:
        if "extra_widget" in kwargs and kwargs["extra_widget"] is not None:
            self.set_extra_widget(kwargs.pop("extra_widget"))

        # Set suggested file name
        if "current_name" in kwargs and kwargs["current_name"] is not None:
            self.set_current_name(kwargs.pop("current_name"))

        # Set suggested folder
        if "current_folder" in kwargs and kwargs["current_folder"] is not None:
            self.set_current_folder(kwargs.pop("current_folder"))

        # Add file filters
        if "filters" in kwargs:
            # Clear old filters:
            for fltr in self.list_filters():
                self.remove_filter(fltr) 
            # Set new filters:
            self.filters = list(kwargs.pop("filters"))
            for fltr in self._get_object_file_filters(self.filters):
                self.add_filter(fltr)

        self.persist = kwargs.pop("persist", self.persist)

        return self

    def _get_object_file_filters(self, filters=[]):
        """ Parses a list of textual file filter globs or Gtk.FileFilter objects
         into Gtk.FileFilter objects """
        for obj in filters:
            if isinstance(obj, Gtk.FileFilter):
                yield obj
            else:
                # if not a Gtk.FileFilter we assume it is a glob tuple
                name, re = obj
                ffilter = Gtk.FileFilter()
                ffilter.set_name(name)
                if isinstance(re, str):
                    ffilter.add_pattern(re)
                else: # if not a single glob, assume an iterable is given
                    for expr in re:
                        ffilter.add_pattern(expr)
                yield ffilter

    #override
    def run(self, *args, **kwargs):
        kwargs['destroy'] = not self.persist
        return run_dialog(self, *args, **kwargs)

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.models.properties import BoolProperty, FloatProperty

from pyxrd.generic.mathtext_support import mt_range
from pyxrd.generic.io import storables
from pyxrd.generic.utils import not_none
from pyxrd.generic.models.properties import InheritableMixin

from pyxrd.refinement.refinables.properties import RefinableMixin

from .base_models import _AbstractProbability
from mvc.models.properties.action_mixins import SetActionMixin

__all__ = [
    "R2G2Model",
    "R2G3Model"
]

@storables.register()
class R2G2Model(_AbstractProbability):
    r"""
    Probability model for Reichweite 2 with 2 components.
    
    The 4 (=g^2) independent variables are:
    
    .. math::
        :nowrap:
    
        \begin{align*}
            & W_1
            & P_{112} (W_1 leq \nicefrac{2}{3})
            \text{ or }P_{211} (W_1 > \nicefrac{2}{3}) \\
            & P_{21}
            & P_{122} (P_{21} leq \nicefrac{1}{2})
            \text{ or }P_{221} (P_{21} > \nicefrac{1}{2}) \\
        \end{align*}
            
    Calculation of the other variables happens as follows:
    
    .. math::
        :nowrap:

        \begin{align*}
            & W_2 = 1 - W_1 \\
            & P_{22} = 1 - P_{21} \\
            & \\
            & W_{21} = W_2 \cdot P_{21} \\
            & W_{21} = W_{12} \\
            & W_{11} = W_1 - W_{21} \\
            & W_{22} = W_{2} \cdot P_{22} \\
            & \\
            & \text{if $W_1 leq \nicefrac{2}{3}$:} \\
            & \quad \text{$P_{112}$ is given}\\
            & \quad P_{211} =
            \begin{dcases}
                \frac{W_{11}}{W_{21}} \cdot P_{112} , & \text{if $W_{21} > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & \\
            & \text{if $W_1 > \nicefrac{2}{3}$:} \\
            & \quad \text{$P_{211}$ is given}\\
            & \quad P_{112} =
            \begin{dcases}
                \frac{W_{21}}{W_{11}} \cdot P_{211} , & \text{if $W_{11} > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & \\
            & P_{212} = 1 - P_{211} \\
            & P_{111} = 1 - P_{112} \\
            & \\
            & \text{if $P_{21} leq \nicefrac{1}{2}$:} \\
            & \quad \text{$P_{122}$ is given}\\
            & \quad P_{221} =
            \begin{dcases}
                \frac{W_{12}}{W_{22}} \cdot P_{122} , & \text{if $W_{22} > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & \\
            & \text{if $P_{21} > \nicefrac{1}{2}$:} \\
            & \quad \text{$P_{221}$ is given}\\
            & \quad P_{122} =
            \begin{dcases}
                \frac{W_{22}}{W_{12}} \cdot P_{221} , & \text{if $W_{12} > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & P_{121} = 1 - P_{122} \\
            & P_{222} = 1 - P_{221} \\
        \end{align*}
    
    """

    # MODEL METADATA:
    class Meta(_AbstractProbability.Meta):
        store_id = "R2G2Model"

    # PROPERTIES:
    _G = 2
    twothirds = 2.0 / 3.0

    inherit_W1 = BoolProperty(
        default=False, text="Inherit flag for W1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    W1 = FloatProperty(
        default=0.75, text="W1 (> 0.5)", math_text=r"$W_1 (> 0.5)$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.5, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_W1", inherit_from="parent.based_on.probabilities.W1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P112_or_P211 = BoolProperty(
        default=False, text="Inherit flag for P112_or_P211",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    P112_or_P211 = FloatProperty(
        default=0.75, text="P112 (W1 < 2/3) or\nP211 (W1 > 2/3)",
        math_text=r"$P_{112} %s$ or $\newlineP_{211} %s$" % (
            mt_range(1.0 / 2.0, "W_1", 2.0 / 3.0),
            mt_range(2.0 / 3.0, "W_1", 1.0)),
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P112_or_P211", inherit_from="parent.based_on.probabilities.P112_or_P211",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P21 = BoolProperty(
        default=False, text="Inherit flag for P21",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    P21 = FloatProperty(
        default=0.75, text="P21", math_text=r"$P_{21}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P21", inherit_from="parent.based_on.probabilities.P21",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P122_or_P221 = BoolProperty(
        default=False, text="Inherit flag for P122_or_P221",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    P122_or_P221 = FloatProperty(
        default=0.75, text="P112 (W1 < 1/2) or\nP221 (W1 > 1/2)",
        math_text=r"$P_{122} %s$ or $\newlineP_{221} %s$" % (
            mt_range(0.0, "W_1", 1.0 / 2.0),
            mt_range(1.0 / 2.0, "W_1", 1.0)),
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P122_or_P221", inherit_from="parent.based_on.probabilities.P122_or_P221",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, W1=0.75, P112_or_P211=0.75, P21=0.75, P122_or_P221=0.75,
            inherit_W1=False, inherit_P112_or_P211=False,
            inherit_P21=False, inherit_P122_or_P221=False, *args, **kwargs):
        super(R2G2Model, self).__init__(R=2, *args, **kwargs)

        with self.data_changed.hold():
            self.W1 = not_none(W1, 0.75)
            self.inherit_W1 = inherit_W1
            self.P112_or_P211 = not_none(P112_or_P211, 0.75)
            self.inherit_P112_or_P211 = inherit_P112_or_P211
            self.P21 = not_none(P21, 0.75)
            self.inherit_P21 = inherit_P21
            self.P122_or_P221 = not_none(P122_or_P221, 0.75)
            self.inherit_P122_or_P221 = inherit_P122_or_P221

            self.update()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update(self):
        with self.monitor_changes():
            self.mW[0] = self.W1
            self.mW[1] = 1.0 - self.mW[0]

            self.mP[1, 0] = self.P21
            self.mP[1, 1] = 1.0 - self.mP[1, 0]

            self.mW[1, 0] = self.mW[1] * self.mP[1, 0]
            self.mW[1, 1] = self.mW[1] * self.mP[1, 1]
            self.mW[0, 1] = self.mW[1, 0]
            self.mW[0, 0] = self.mW[0] - self.mW[1, 0]

            if self.mW[0] <= self.twothirds:
                self.mP[0, 0, 1] = self.P112_or_P211
                if self.mW[1, 0] == 0.0:
                    self.mP[1, 0, 0] = 0.0
                else:
                    self.mP[1, 0, 0] = self.mP[0, 0, 1] * self.mW[0, 0] / self.mW[1, 0]
            else:
                self.mP[1, 0, 0] = self.P112_or_P211
                if self.mW[0, 0] == 0.0:
                    self.mP[0, 0, 1] = 0.0
                else:
                    self.mP[0, 0, 1] = self.mP[1, 0, 0] * self.mW[1, 0] / self.mW[0, 0]
            self.mP[1, 0, 1] = 1.0 - self.mP[1, 0, 0]
            self.mP[0, 0, 0] = 1.0 - self.mP[0, 0, 1]

            if self.mP[1, 0] <= 0.5:
                self.mP[0, 1, 1] = self.P122_or_P221
                self.mP[1, 1, 0] = self.mP[0, 1, 1] * self.mW[0, 1] / self.mW[1, 1]
            else:
                self.mP[1, 1, 0] = self.P122_or_P221
                self.mP[0, 1, 1] = self.mP[1, 1, 0] * self.mW[1, 1] / self.mW[0, 1]
            self.mP[0, 1, 0] = 1.0 - self.mP[0, 1, 1]
            self.mP[1, 1, 1] = 1.0 - self.mP[1, 1, 0]

            self.solve()
            self.validate()

    pass # end of class

@storables.register()
class R2G3Model(_AbstractProbability):
    r"""
    
    (Restricted) probability model for Reichweite 2 with 3 components.
    
    The (due to restrictions only) 6 independent variables are:
    
    .. math::
        :nowrap:

        \begin{align*}
            & W_{1}
            & P_{111} \text{(if $\nicefrac{1}{2} \leq W_1 < \nicefrac{2}{3}$) or} P_{x1x} \text{(if $\nicefrac{2}{3} \leq W_1 \leq 1)$ with $x \in \left\{ {2,3} \right\}$} \\
            & G_1 = \frac{W_2}{W_2 + W_3}
            & G_2 = \frac{W_{212} + W_{213}}{W_{212} + W_{213} + W_{312} + W_{313}} \\
            & G_3 = \frac{W_{212}}{W_{212} + W_{213}}
            & G_4 = \frac{W_{312}}{W_{312} + W_{313}} \\
        \end{align*}
        
    This model can not describe mixed layers in which the last two components
    occur right after each other in a stack. In other words there is always
    an alternation between (one or more) layers of the first component and a 
    single layer of the second or third component. Therefore, the weight 
    fraction of the first component (:math:`W_1`) needs to be > than 1/2.
    
    The restriction also translates in the following:
    
    .. math::
        :nowrap:
        
        \begin{align*}
            & P_{22} = P_{23} = P_{32} = P_{33} = 0 \\
            & P_{21} = P_{31} = 1 \\
            & \\
            & P_{122} = P_{123} = P_{132} = P_{133} = 0 \\
            & P_{121} = P_{131} = 1 \\
            & \\
            & P_{222} = P_{223} = P_{232} = P_{233} = 0 \\
            & P_{221} = P_{231} = 1 \\
            & \\
            & P_{322} = P_{323} = P_{332} = P_{333} = 0 \\
            & P_{321} = P_{331} = 1 \\
        \end{align*}
    
    Using the above, we can calculate a lot of the weight fractions of stacks:
    
    .. math::
        :nowrap:
    
        \begin{align*}
            & W_{22} = W_{23} = W_{32} = W_{33} 0 \\
            & W_{21} = W_{2} \\
            & W_{31} = W_{3} \\
            & \\
            & W_{122} = W_{123} = W_{132} = W_{133} = 0 \\
            & W_{121} = W_{12} = W_{21} = W_2 \\
            & W_{131} = W_{13} = W_{31} = W_3 \\
            & W_{11} = W_1 - W_{12} - W_{13} \\
            & \\
            & W_{221} = W_{231} = W_{222} = W_{223} = W_{232} = W_{233} = 0 \\
            & W_{331} = W_{331} = W_{322} = W_{323} = W_{332} = W_{333} = 0 \\             
        \end{align*}

    Then the remaining fractions and probablities can be calculated as follows:
    
    .. math::
        :nowrap:
    
        \begin{align*}
            & W_2 = G_1 * (1 - W_1) \\
            & W_3 = 1 - W_1 - W_2 \\
            & \\
            & W_x = W_2 + W_3 &
            & \text{if $W_1 < \nicefrac{2}{3}$:} \\
            & \quad \text{$P_{111}$ is given}\\
            & \quad P_{x1x} = 
            \begin{dcases}
                1 - \frac{W_1 - W_x}{W_x} \cdot (1 - P_{111}, & \text{if $W_x > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\ 
            & \\
            & \text{if $W_1 \geq \nicefrac{2}{3}$:} \\
            & \quad \text{$P_{x1x}$ is given}\\
            & \quad P_{111} = 
            \begin{dcases}
                1 - \frac{W_x}{W_1 - W_x} \cdot (1 - P_{x1x}, & \text{if $(W_1 - W_x) > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & \\
            & W_{x1x} = W_x \cdot P_{x1x} \\
            & W_{21x} = G_2 \cdot W_{x1x} \\
            & W_{31x} = W_{x1x} - W_{21x} \\
            & \\
            & W_{212} = G_3 \cdot W_{21x} \\
            & W_{213} = (1 - G_3) \cdot W_{21x} \\
            & W_{211} = W_{21} - W_{212} - W_{213} \\
            & \\
            & W_{312} = G_4 \cdot W_{31x} \\
            & W_{313} = (1 - G_4) \cdot W_{31x} \\
            & W_{311} = W_{31} - W_{312} - W_{313} \\
            & \\
            & W_{111} = W_{11} \cdot P_{111} \\
            & W_{112} = W_{12} - W_{212} - W_{312} \\
            & W_{112} = W_{13} - W_{213} - W_{313} \\
            & \\
            & \text{Calculate the remaining P using:} \\
            & P_{ijk} = 
            \begin{dcases}
                \frac{W_{ijk}}{W_{ij}}, & \text{if $W_{ij} > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\ 
        \end{align*}
        
    """

    # MODEL METADATA:
    class Meta(_AbstractProbability.Meta):
        store_id = "R2G3Model"

    # PROPERTIES:
    _G = 3
    twothirds = 2.0 / 3.0

    inherit_W1 = BoolProperty(
        default=False, text="Inherit flag for W1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    W1 = FloatProperty(
        default=0.8, text="W1 (> 0.5)", math_text=r"$W_1 (> 0.5)$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.5, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_W1", inherit_from="parent.based_on.probabilities.W1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P111_or_P212 = BoolProperty(
        default=False, text="Inherit flag for P112_or_P211",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    P111_or_P212 = FloatProperty(
        default=0.9, text="P111 (W1 < 2/3) or\nPx1x (W1 > 2/3)",
        math_text=r"$P_{111} %s$ or $\newline P_{x1x} %s$" % (
            mt_range(0.5, "W_1", 2.0 / 3.0),
            mt_range(2.0 / 3.0, "W_1", 1.0)),
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P111_or_P212", inherit_from="parent.based_on.probabilities.P111_or_P212",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G1 = BoolProperty(
        default=False, text="Inherit flag for G1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G1 = FloatProperty(
        default=0.9, text="W2/(W2+W3)",
        math_text=r"$\large\frac{W_2}{W_3 + W_2}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G1", inherit_from="parent.based_on.probabilities.G1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G2 = BoolProperty(
        default=False, text="Inherit flag for G2",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G2 = FloatProperty(
        default=0.9, text="(W212+W213)/(W212+W213+W312+W313)",
        math_text=r"$\large\frac{W_{212} + W_{213}}{W_{212} + W_{213} + W_{312} + W_{313}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G2", inherit_from="parent.based_on.probabilities.G2",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G3 = BoolProperty(
        default=False, text="Inherit flag for G3",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G3 = FloatProperty(
        default=0.9, text="W212/(W212+W213)",
        math_text=r"$\large\frac{W_{212}}{W_{212} + W_{213}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G3", inherit_from="parent.based_on.probabilities.G3",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G4 = BoolProperty(
        default=False, text="Inherit flag for G4",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G4 = FloatProperty(
        default=0.9, text="W312/(W312+W313)",
        math_text=r"$\large\frac{W_{312}}{W_{312} + W_{313}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G4", inherit_from="parent.based_on.probabilities.G4",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, W1=0.8, P111_or_P212=0.9, G1=0.9, G2=0.9, G3=0.9, G4=0.9,
        inherit_W1=False, inherit_P111_or_P212=False, inherit_G1=False,
        inherit_G2=False, inherit_G3=False, inherit_G4=False, *args, **kwargs):
        super(R2G3Model, self).__init__(R=2, *args, **kwargs)

        with self.data_changed.hold():
            self.W1 = not_none(W1, 0.8)
            self.inherit_W1 = inherit_W1
            self.P111_or_P212 = not_none(P111_or_P212, 0.9)
            self.inherit_P111_or_P212 = inherit_P111_or_P212
            self.G1 = not_none(G1, 0.9)
            self.inherit_G1 = inherit_G1
            self.G2 = not_none(G2, 0.9)
            self.inherit_G2 = inherit_G2
            self.G3 = not_none(G3, 0.9)
            self.inherit_G3 = inherit_G3
            self.G4 = not_none(G4, 0.9)
            self.inherit_G4 = inherit_G4

            self.update()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update(self):
        with self.monitor_changes():

            # calculate Wx's:
            self.mW[0] = self.W1
            self.mW[1] = (1.0 - self.mW[0]) * self.G1
            self.mW[2] = 1.0 - self.mW[0] - self.mW[1]

            # consequences of restrictions:
            self.mW[1, 1] = 0
            self.mW[1, 2] = 0
            self.mW[2, 1] = 0
            self.mW[2, 2] = 0
            self.mW[0, 1, 0] = self.mW[0, 1] = self.mW[1, 0] = self.mW[1]
            self.mW[0, 2, 0] = self.mW[0, 2] = self.mW[2, 0] = self.mW[2]
            self.mW[0, 0] = self.mW[0] - self.mW[0, 1] - self.mW[0, 2]

            # continue calculations:
            Wx = self.mW[1] + self.mW[2]
            if self.mW[0] < self.twothirds:
                self.mP[0, 0, 0] = self.P111_or_P212
                Px0x = 1 - (self.mW[0] - Wx) / Wx * (1 - self.mP[0, 0, 0]) if Wx != 0 else 0.0
            else:
                Px0x = self.P111_or_P212
                self.mP[0, 0, 0] = 1 - Wx / (self.mW[0] - Wx) * (1 - Px0x) if (self.mW[0] - Wx) != 0 else 0.0

            Wx0x = Wx * Px0x
            W10x = self.G2 * Wx0x
            W20x = Wx0x - W10x

            self.mW[1, 0, 1] = self.G3 * W10x
            self.mW[1, 0, 2] = (1 - self.G3) * W10x
            self.mW[1, 0, 0] = self.mW[1, 0] - self.mW[1, 0, 1] - self.mW[1, 0, 2]

            self.mW[2, 0, 1] = self.G4 * W20x
            self.mW[2, 0, 2] = (1 - self.G4) * W20x
            self.mW[2, 0, 0] = self.mW[2, 0] - self.mW[2, 0, 1] - self.mW[2, 0, 2]

            self.mW[0, 0, 0] = self.mW[0, 0] * self.mP[0, 0, 0]
            self.mW[0, 0, 1] = self.mW[0, 1] - self.mW[1, 0, 1] - self.mW[2, 0, 1]
            self.mW[0, 0, 2] = self.mW[0, 2] - self.mW[1, 0, 2] - self.mW[2, 0, 2]

            # Calculate remaining P:
            for i in range(3):
                for j in range(3):
                    for k in range(3):
                        self.mP[i, j, k] = self.mW[i, j, k] / self.mW[i, j] if self.mW[i, j] > 0 else 0.0

            self.solve()
            self.validate()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from functools import wraps

class StatusBarMixin(object):
    @property
    def statusbar(self):
        if self.parent is not None:
            return self.parent.statusbar
        elif self.view is not None:
            return self.view['statusbar']
        else:
            return None

    @property
    def status_cid(self):
        if self.statusbar is not None:
            return self.statusbar.get_context_id(self.__class__.__name__)
        else:
            return None

    @staticmethod
    def status_message(message, cid=None):
        def decorator(func):
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                self.push_status_msg(message, cid)
                res = func(self, *args, **kwargs)
                self.pop_status_msg(cid)
                return res
            return wrapper
        return decorator

    def push_status_msg(self, msg, cid=None):
        if cid is not None:
            cid = self.statusbar.get_context_id(cid)
        else:
            cid = self.status_cid
        if cid is not None:
            self.statusbar.push(cid, msg)

    def pop_status_msg(self, cid=None):
        if cid is not None:
            cid = self.statusbar.get_context_id(cid)
        else:
            cid = self.status_cid
        if cid is not None:
            self.statusbar.pop(cid)

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon

class DummyAsyncResult(object):
    """ A non-asynchronous dummy implementation of the AsyncResult object """
    def __init__(self, func):
        self.result = func()
        
    def get(self):
        return self.result
    
    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, struct
from io import SEEK_SET, SEEK_CUR

import numpy as np

from pyxrd.generic.io.utils import get_case_insensitive_glob
from pyxrd.generic.utils import u

from ..base_parser import BaseParser
from .namespace import xrd_parsers
from .xrd_parser_mixin import XRDParserMixin

@xrd_parsers.register_parser()
class BrkRAWParser(XRDParserMixin, BaseParser):
    """
        Bruker *.RAW format parser
    """

    description = "Bruker/Siemens Binary V1/V2/V3 *.RAW"
    extensions = get_case_insensitive_glob("*.RAW")
    mimetypes = ["application/octet-stream", ]

    __file_mode__ = "rb"

    @classmethod
    def _clean_bin_str(cls, val):
        return u(val.replace("\0".encode(), "".encode()).strip())

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False):
        f = fp

        try:
            basename = u(os.path.basename(filename))
        except:
            basename = None

        # Go to the start of the file
        f.seek(0, SEEK_SET)

        # Read file format version:
        version = f.read(4).decode("utf-8")
        if version == "RAW ":                             version = "RAW1"
        elif version == "RAW2":                           version = "RAW2"
        elif version == "RAW1" and str(f.read(3)) == ".01": version = "RAW3"

        if version == "RAW1":

            # This format does not allow getting the exact number of samples,
            # so start with one and append where needed:
            isfollowed = 1
            num_samples = 0
            while isfollowed > 0:

                twotheta_count = int(struct.unpack("I", f.read(4))[0])
                # Check if this is an early "RAW " formatted file where the
                # "RAW " is repeated for each sample:
                if num_samples > 0 and twotheta_count == int(struct.unpack("I", "RAW ")[0]):
                    twotheta_count = int(struct.unpack("I", f.read(4))[0])

                # Step counting time, 2-theta step size and scanning mode:
                time_step, twotheta_step, scan_mode = struct.unpack("fff", f.read(12)) #@UnusedVariable
                # Skip 4 bytes, and read 2-theta starting position:
                f.seek(4, SEEK_CUR)
                twotheta_min, = struct.unpack("f", f.read(4))
                twotheta_max = twotheta_min + twotheta_step * float(twotheta_count)
                # Skip 12 bytes
                # (contain theta, khi and phi start point for eularian craddles)
                f.seek(12, SEEK_CUR)
                # Read sample name and wavelengths:
                sample_name = cls._clean_bin_str(f.read(32))
                alpha1, alpha2 = struct.unpack("ff", f.read(8))
                # Skip 72 bytes:
                f.seek(72, SEEK_CUR)
                isfollowed, = struct.unpack("I", f.read(4))

                # Get data position and skip for now:
                data_start = f.tell()
                f.seek(twotheta_count * 4, SEEK_CUR)

                # Adapt XRDFile list
                data_objects = cls._adapt_data_object_list(
                    data_objects,
                    num_samples=(num_samples + 1),
                    only_extend=True
                )

                data_objects[num_samples].update(
                    filename=basename,
                    version=version,
                    name=sample_name,
                    time_step=time_step,
                    twotheta_min=twotheta_min,
                    twotheta_max=twotheta_max,
                    twotheta_step=twotheta_step,
                    twotheta_count=twotheta_count,
                    alpha1=alpha1,
                    alpha2=alpha2,
                    data_start=data_start
                )

                num_samples += 1

        elif version == "RAW2":

            # Read number of sample ranges:
            num_samples, = struct.unpack("H", f.read(2))

            # Adapt XRDFile list
            data_objects = cls._adapt_data_object_list(data_objects, num_samples=num_samples)

            # Read sample name:
            f.seek(8, SEEK_SET)
            sample_name = cls._clean_bin_str(f.read(32))
            # Meta-data description, skip for now:
            # description = u(str(f.read(128)).replace("\0", "").strip())
            # date = u(str(f.read(10)).replace("\0", "").strip())
            # time = u(str(f.read(5)).replace("\0", "").strip())

            # Read wavelength information:
            f.seek(148, SEEK_CUR)
            target_type = u(str(f.read(2)).replace("\0", "").strip()) #@UnusedVariable
            alpha1, alpha2, alpha_factor = struct.unpack("fff", f.read(12))

            # Total runtime in seconds: (not used fttb)
            f.seek(8, SEEK_CUR)
            time_total, = struct.unpack("f", f.read(4)) #@UnusedVariable

            # Move to first sample header start:
            f.seek(256, SEEK_SET)

            # Read in per-sample meta data
            for i in range(num_samples):
                header_start = f.tell()
                header_length, twotheta_count = struct.unpack("HH", f.read(4))
                data_start = header_start + header_length

                # Read step size and start angle:
                f.seek(header_start + 12) # = 256 + 4 + 8 skipped bytes
                twotheta_step, twotheta_min = struct.unpack("ff", f.read(8))
                twotheta_max = twotheta_min + twotheta_step * float(twotheta_count)

                # Read up to end of data:
                f.seek(data_start + twotheta_count * 4, SEEK_SET)

                # Update XRDFile object:
                data_objects[i].update(
                    filename=basename,
                    version=version,
                    name=sample_name,
                    twotheta_min=twotheta_min,
                    twotheta_max=twotheta_max,
                    twotheta_step=twotheta_step,
                    twotheta_count=twotheta_count,
                    alpha1=alpha1,
                    alpha2=alpha2,
                    alpha_factor=alpha_factor,
                    data_start=data_start
                )

        elif version == "RAW3":

            # Read file status:    
            f.seek(8, SEEK_SET)       
            file_status = { #@UnusedVariable
                1: "done",
                2: "active",
                3: "aborted",
                4: "interrupted"
            }[int(struct.unpack("I", f.read(4))[0])]

            # Read number of samples inside this file:
            f.seek(12, SEEK_SET)
            num_samples, = struct.unpack("I", f.read(4))

            # Read in sample name:
            f.seek(326, SEEK_SET)
            sample_name =  cls._clean_bin_str(f.read(60))

            # Goniometer radius:
            f.seek(564, SEEK_SET)
            radius = float(struct.unpack("f", f.read(4))[0])

            # Fixed divergence:
            f.seek(568, SEEK_SET)
            divergence = float(struct.unpack("f", f.read(4))[0])

            # Primary soller
            f.seek(576, SEEK_SET)
            soller1 = float(struct.unpack("f", f.read(4))[0])

            # Secondary soller
            f.seek(592, SEEK_SET)
            soller2 = float(struct.unpack("f", f.read(4))[0])

            # Get anode type:
            f.seek(608, SEEK_SET)
            target_type = str(f.read(4)) #@UnusedVariable

            # Get wavelength info:
            f.seek(616, SEEK_SET)
            alpha_average, alpha1, alpha2, beta, alpha_factor = (#@UnusedVariable
                struct.unpack("ddddd", f.read(8 * 5)))

            # Get total recording time:
            f.seek(664, SEEK_SET)
            time_total, = struct.unpack("f", f.read(4)) #@UnusedVariable
            
            # Adapt XRDFile lis & Skip to first block:t
            data_objects = cls._adapt_data_object_list(data_objects, num_samples=num_samples)
            f.seek(712, SEEK_SET)
            
            # Read in per-sample meta data
            for i in range(num_samples):
                # Store the start of the header:
                header_start = f.tell()

                # Get header length
                f.seek(header_start + 0, SEEK_SET)
                header_length, = struct.unpack("I", f.read(4))
                assert header_length == 304, "Invalid format!"
                
                # Get data count and
                f.seek(header_start + 4, SEEK_SET)
                twotheta_count, = struct.unpack("I", f.read(4))
                
                # Get theta start positions
                f.seek(header_start + 8, SEEK_SET)
                theta_min, twotheta_min = struct.unpack("dd", f.read(8 * 2))#@UnusedVariable               

                # Read step size
                f.seek(header_start + 176, SEEK_SET)
                twotheta_step, = struct.unpack("d", f.read(8))

                # Read counting time
                f.seek(header_start + 192, SEEK_SET)
                time_step, = struct.unpack("d", f.read(8))

                # Read the used wavelength
                f.seek(header_start + 240, SEEK_SET)
                alpha_used, = struct.unpack("d", f.read(8))#@UnusedVariable

                # Supplementary header size:
                f.seek(header_start + 256, SEEK_SET)
                supp_headers_size, = struct.unpack("I", f.read(4))
                data_start = header_start + header_length + supp_headers_size

                # Move to the end of the data:
                f.seek(data_start + twotheta_count * 4)
                
                # Calculate last data point
                twotheta_max = twotheta_min + twotheta_step * float(twotheta_count - 0.5)
                
                data_objects[i].update(
                    filename=basename,
                    version=version,
                    name=sample_name,
                    twotheta_min=twotheta_min,
                    twotheta_max=twotheta_max,
                    twotheta_step=twotheta_step,
                    twotheta_count=twotheta_count,
                    alpha1=alpha1,
                    alpha2=alpha2,
                    alpha_factor=alpha_factor,
                    data_start=data_start,
                    radius=radius,
                    soller1=soller1,
                    soller2=soller2,
                    divergence=divergence
                )

        else:
            raise IOError("Only verson 1, 2 and 3 *.RAW files are supported!")

        if close: f.close()
        return data_objects

    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=False):
        for data_object in data_objects:
            if data_object.data == None:
                data_object.data = []
            
            # Parse data:
            if fp is not None:
                if data_object.version in ("RAW1", "RAW2", "RAW3"):
                    fp.seek(data_object.data_start)
                    n = 0
                    while n < data_object.twotheta_count:
                        y, = struct.unpack("f", fp.read(4))
                        x = data_object.twotheta_min + data_object.twotheta_step * float(n + 0.5)
                        data_object.data.append([x,y])
                        n += 1
                else:
                    raise IOError("Only verson 1, 2 and 3 *.RAW files are supported!")


            data_object.data = np.array(data_object.data)

        if close: fp.close()
        return data_objects

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .base import AbstractTestProbModel

from test.utils import create_object_attribute_test

from pyxrd.probabilities.models.R3models import *
import unittest

__all__ = [
    'TestR3G2Model'
]

class TestR3G2Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R3G2Model

    test_W1 = create_object_attribute_test("prob_model", "W1", 0.7)
    test_P1111_or_P2112 = create_object_attribute_test("prob_model", "P1111_or_P2112", 0.7)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, types
from zipfile import ZipFile

from pyxrd.generic.io.utils import get_case_insensitive_glob
from pyxrd.generic.utils import not_none

from ..base_parser import BaseParser
from ..xml_parser_mixin import XMLParserMixin
from .xrd_parser_mixin import XRDParserMixin
from .namespace import xrd_parsers

@xrd_parsers.register_parser()
class BrkBRMLParser(XRDParserMixin, XMLParserMixin, BaseParser):
    """
        Bruker *.BRML format parser
    """

    description = "Bruker BRML files *.BRML"
    extensions = get_case_insensitive_glob("*.BRML")
    mimetypes = ["application/zip", ]

    __file_mode__ = "r"

    @classmethod
    def _get_file(cls, fp, close=None):
        """
            Returns a three-tuple:
            filename, zipfile-object, close
        """
        if isinstance(fp, str):
            return fp, ZipFile(fp, cls.__file_mode__), True if close is None else close
        else:
            return getattr(fp, 'name', None), ZipFile(fp, cls.__file_mode__), False if close is None else close

    @classmethod
    def _get_raw_data_files(cls, f, folder):
        """
            Processes DataContainer.xml and returns a list of xml raw data
            filepaths and the sample name
        """
        
        contf = f.open(r"%s/DataContainer.xml" % folder, cls.__file_mode__)
        data = contf.read()
        contf.close()
        
        _, root = cls.get_xml_for_string(data)
        sample_name = root.find("./MeasurementInfo").get("SampleName")

        raw_data_files = []
        for child in root.find("./RawDataReferenceList"):
            raw_data_files.append(child.text)
            
        return raw_data_files, sample_name

    @classmethod
    def _get_header_dict(cls, f, folder):
        header_d = {}

        contf = f.open(r"%s/MeasurementContainer.xml" % folder, cls.__file_mode__)
        data = contf.read()
        contf.close()
        
        _, root = cls.get_xml_for_string(data)

        radius_path = "./HardwareLogicExt/Instrument/BeamPathContainers" + \
               "/BeamPathContainerAbc[@VisibleName='PrimaryTrack']/%s"

        tube_path = "./HardwareLogicExt/Instrument/BeamPathContainers" + \
               "/BeamPathContainerAbc[@VisibleName='PrimaryTrack']/BankPositions/" + \
               "BankPosition/MountedComponent/MountedTube/%s"

        soller1_path = "./HardwareLogicExt/Instrument/BeamPathContainers" + \
               "/BeamPathContainerAbc[@VisibleName='PrimaryTrack']/BankPositions/" + \
               "BankPosition/MountedComponent[@VisibleName='SollerMount']/%s"

        soller2_path = "./HardwareLogicExt/Instrument/BeamPathContainers" + \
               "/BeamPathContainerAbc[@VisibleName='SecondaryTrack']/BankPositions/" + \
               "BankPosition/MountedComponent[@VisibleName='SollerMount']/%s"

        divergence_path = "./HardwareLogicExt/Instrument/BeamPathContainers" + \
               "/BeamPathContainerAbc[@VisibleName='PrimaryTrack']/BankPositions/" + \
               "BankPosition/MountedComponent/Restrictions[@FieldName='OpeningDegree']/%s"

        header_d.update(
            alpha1=float(cls.get_val(root, tube_path % "WaveLengthAlpha1", "Value")) / 10,
            alpha2=float(cls.get_val(root, tube_path % "WaveLengthAlpha2", "Value")) / 10,
            alpha_average=float(cls.get_val(root, tube_path % "WaveLengthAverage", "Value")) / 10,
            beta=float(cls.get_val(root, tube_path % "WaveLengthBeta", "Value")) / 10,
            alpha_factor=cls.get_val(root, tube_path % "WaveLengthRatio", "Value"),
            target_type=cls.get_val(root, tube_path % "TubeMaterial", "Value"),
            soller1=cls.get_val(root, soller1_path % "Deflection", "Value"),
            soller2=cls.get_val(root, soller2_path % "Deflection", "Value"),
            radius=float(cls.get_val(root, radius_path % "Radius", "Value", 0)) / 10.0, #convert to cm
            divergence=cls.get_val(root, divergence_path % "Data", "Value")
        )
        
        return header_d

    @classmethod
    def parse(cls, fp, data_objects=None, close=False):
        filename, fp, close = cls._get_file(fp, close=close)

        try:
            basename = os.path.basename(filename)
        except AttributeError:
            basename = None

        num_samples = 0

        zipinfos = fp.infolist()

        processed_folders = []

        data_objects = not_none(data_objects, [])

        for zipinfo in zipinfos:
            if zipinfo.filename.count('/') == 1 and "DataContainer.xml" in zipinfo.filename:

                folder = os.path.dirname(zipinfo.filename)
                if not folder in processed_folders:

                    processed_folders.append(folder)

                    header_d = cls._get_header_dict(fp, folder)
                    raw_data_files, sample_name = cls._get_raw_data_files(fp, folder)
               
                    for raw_data_filename in raw_data_files:
                        contf = fp.open(raw_data_filename)
                        
                        _, root = cls.get_xml_for_file(contf)
                        
                        isScan = not ("NonAmbientModeData" in root.find("./DataRoutes/DataRoute/ScanInformation").get("ScanName"))

                        if isScan:
                            for route in root.findall("./DataRoutes/DataRoute"):
                                # Adapt XRDFile list & get last addition:
                                data_objects = cls._adapt_data_object_list(
                                    data_objects,
                                    num_samples=(num_samples + 1),
                                    only_extend=True
                                )
                                data_object = data_objects[num_samples]
    
                                # Get the Datum tags:
                                datums = route.findall("Datum")
                                data = []
    
                                # Parse the RawDataView tags to find out what index in
                                # the datum is used for what type of data:
                                enabled_datum_index = None
                                twotheta_datum_index = None
                                intensity_datum_index = None
                                steptime_datum_index = None
                                relative_humidity_data, relative_humidity_index = None, None
                                temperature_data, temperature_index = None, None
                                temperature_index = None
                                for dataview in route.findall("./DataViews/RawDataView"):
                                    index = int(dataview.get("Start", 0))
                                    name = dataview.get("LogicName") or "Undefined"
                                    xsi_type = dataview.get("{http://www.w3.org/2001/XMLSchema-instance}type") or "Undefined"
                                    if name == "MeasuredTime":
                                        steptime_datum_index = index
                                    elif name == "AbsorptionFactor":
                                        enabled_datum_index = index
                                    elif name == "Undefined" and xsi_type == "VaryingRawDataView":
                                        for i, definition in enumerate(dataview.findall("./Varying/FieldDefinitions")):
                                            if definition.get("TwoTheta"):
                                                index += i
                                                break
                                        twotheta_datum_index = index
                                    elif name == "Undefined" and xsi_type == "RecordedRawDataView":
                                        logic_name = dataview.find("Recording").get("LogicName")
                                        if logic_name == "ScanCounter":
                                            intensity_datum_index = index
                                        elif logic_name == "modeActualHum":
                                            relative_humidity_index = index
                                            relative_humidity_data = []
                                        elif logic_name == "modeActualTemp":
                                            temperature_index = index
                                            temperature_data = []
                                            
                                # Parse the SubScanInfo list (usually only one), and
                                # then parse the datums accordingly
                                twotheta_min = None
                                twotheta_max = None
                                twotheta_count = 0
                                for subscan in route.findall("./SubScans/SubScanInfo"):
                                    # Get the steps, where to start and the planned
                                    # time per step (measuredTimePerStep deviates
                                    # if the recording was interrupted):
                                    steps = int(subscan.get("MeasuredSteps"))
                                    start = int(subscan.get("StartStepNo"))
                                    steptime = float(subscan.get("PlannedTimePerStep"))
    
                                    for datum in datums[start:start + steps]:
                                        values = datum.text.split(",")
                                        if values[enabled_datum_index] == "1":
                                            # Fetch values from the list:
                                            datum_steptime = float(values[steptime_datum_index])
                                            intensity = float(values[intensity_datum_index])
                                            intensity /= float(steptime * datum_steptime)
                                            twotheta = float(values[twotheta_datum_index])
                                            
                                            # If we have temperature or RH data, get them as well:
                                            if temperature_index is not None:
                                                temperature = float(values[temperature_index])
                                                temperature_data.append(temperature) 
                                            if relative_humidity_index is not None:
                                                relative_humidity = float(values[relative_humidity_index])
                                                relative_humidity_data.append(relative_humidity)                                        
    
                                            # Keep track of min 2theta:
                                            if twotheta_min is None:
                                                twotheta_min = twotheta
                                            else:
                                                twotheta_min = min(twotheta_min, twotheta)
    
                                            # Keep track of max 2theta:
                                            if twotheta_max is None:
                                                twotheta_max = twotheta
                                            else:
                                                twotheta_max = min(twotheta_max, twotheta)
    
                                            # Append point and increase count:
                                            data.append([twotheta, intensity])
                                            twotheta_count += 1
    
                                #Update header:
                                data_object.update(
                                    filename=basename,
                                    name=sample_name,
                                    time_step=1, # we converted to CPS
                                    twotheta_min=twotheta_min,
                                    twotheta_max=twotheta_max,
                                    twotheta_count=twotheta_count,
                                    **header_d
                                )
    
                                data_object.data = data
                                
                                # These might be None:
                                data_object.temperature_data = temperature_data
                                data_object.relative_humidity_data = relative_humidity_data
    
                                num_samples += 1

                        #end for
                        contf.close()
                        
                    #end for
                #end if
            #end if
        #end for

        if close: fp.close()
        return data_objects

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.models.properties import FloatProperty

class ProbabilityProperty(FloatProperty):
    """
     A descriptor that will invoke the 'update' method on the instance
     it belongs to.
    """

    def __init__(self, clamp=False, **kwargs):
        super(ProbabilityProperty, self).__init__(**kwargs)

    def __set__(self, instance, value):
        super(ProbabilityProperty, self).__set__(instance, value)
        instance.update()

    pass # end of class



from .storable_xy_data import StorableXYData
from .pyxrd_line import PyXRDLine
from .calculated_line import CalculatedLine
from .experimental_line import ExperimentalLine

__all__ = [
    "StorableXYData",
    "PyXRDLine",
    "CalculatedLine",
    "ExperimentalLine"
]


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Tobias Weber
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import threading

from .observable import Observable

class Signal(Observable):
    """
        A Signal that can either:
         - be 'held' from firing until a code block has finished, e.g. to
           prevent numerous events from firing, when one final event would
           be enough
         - be stopped from firing altogether, even after the code block has
           finished 
        
        Holding signals back:
         ...
         object.hold_signal = HoldableSignal()
         ...
         with object.hold_signal.hold():
             # this code block can call emit() on the hold_signal but it will not
             # actually emit the signal until the 'with' block is left
         ...
         
        Ignoring signals:
         ...
         object.hold_signal =  HoldableSignal()
         ...
         with object.hold_signal.ignore():
             # this code block can call emit() on the hold_signal but it will not
             # actually emit the signal, even after leaving the 'with' block
    """
    # PROPERTIES:
    clock = threading.RLock()

    # INIT:
    def __init__(self, *args, **kwargs):
        super(Signal, self).__init__(*args, **kwargs)
        self._counter = 0
        self._emissions_pending = False
        self._ignore_levels = []

    # WITH FUNCTIONALITY:
    def __enter__(self):
        with self.clock:
            self._counter += 1;

    def __exit__(self, *args):
        with self.clock:
            self._counter -= 1;
            if self._counter < 0:
                raise RuntimeError("Negative counter in CounterLock object! Did you call __exit__ too many times?")
            if len(self._ignore_levels) > 0 and self._counter == self._ignore_levels[-1]:
                self._ignore_levels.pop()
            elif self._counter == 0 and self._emissions_pending:
                # Fire the signal when we leave the with block
                self.emit()

    def hold(self):
        return self

    def ignore(self):
        self._ignore_levels.append(self._counter)
        return self

    def hold_and_emit(self):
        self._emit_pending() # set our pending flag
        return self

    def emit(self, arg=None):
        if self._counter == 0:
            self._emissions_pending = False
            for model, name in self.__get_models__():
                model.notify_signal_emit(name, arg)
                pass
        else:
            self._emit_pending()

    def _emit_pending(self):
        if not self._emissions_pending:
            self._emissions_pending = bool(len(self._ignore_levels) == 0)

    pass #end of class


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from traceback import print_exc

from . import settings
import Pyro4

if __name__ == "__main__":
    try:
        server = Pyro4.Proxy("PYRONAME:%s" % settings.PYRO_NAME)
        server.shutdown()
    except:
        logging.error("Error when trying to shut down Pyro server!")
        print_exc()


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .phase import Phase
from .raw_pattern_phase import RawPatternPhase
from .component import Component

__all__ = [
    "Phase",
    "RawPatternPhase",
    "Component",
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

class RefineMethodOption(object):
    """ Descriptor for refinement methods """
    # TODO MERGE THIS WITH THE DESCRIPTORS FROM THE OTHER BRANCH

    label = None

    def __init__(self, description, default=None, limits=[None, None], value_type=object, fget=None, fset=None, fdel=None, doc=None, label=None):
        self.description = description
        self.limits = limits
        self.value_type = value_type

        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

        self.label = label
        self.default = default

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.fget is None:
            return getattr(instance, "_%s" % self.label, self.default)
        else:
            return self.fget(instance)

    def __set__(self, instance, value):
        _min, _max = self.limits
        if self.value_type in (str, int, float): value = self.value_type(value)
        if _min is not None: value = max(value, _min)
        if _max is not None: value = min(value, _max)

        if self.fset is None:
            setattr(instance, "_%s" % self.label, value)
        else:
            return self.fset(instance, value)

    def __delete__(self, instance):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(instance)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, locale
import logging
logger = logging.getLogger(__name__)

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
from mvc.adapters.gtk_support.tree_view_adapters import wrap_xydata_to_treemodel
from mvc.adapters import DummyAdapter

from pyxrd.generic.controllers import BaseController, DialogController, TreeViewMixin
from pyxrd.generic.views.treeview_tools import setup_treeview, new_text_column
from pyxrd.file_parsers.xrd_parsers import xrd_parsers
from pyxrd.file_parsers.exc_parsers import exc_parsers

from pyxrd.goniometer.controllers import InlineGoniometerController

from pyxrd.generic.controllers.line_controllers import (
    LinePropertiesController,
    BackgroundController,
    SmoothDataController,
    AddNoiseController,
    ShiftDataController,
    StripPeakController,
    CalculatePeakPropertiesController
)

from pyxrd.generic.views.line_views import (
    BackgroundView,
    SmoothDataView,
    AddNoiseView,
    ShiftDataView,
    StripPeakView,
    CalculatePeakPropertiesView
)

class SpecimenController(DialogController, TreeViewMixin):
    """
        Specimen controller.
    """

    widget_handlers = {
        'custom':  'custom_handler',
    }

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    @staticmethod
    def custom_handler(self, prop, widget):
        if prop.label in ("goniometer"):
            self.gonio_ctrl = InlineGoniometerController(view=self.view.gonio_view, model=self.model.goniometer, parent=self)
            ad = DummyAdapter(controller=self, prop=prop) # TODO FIXME
            return ad

    def setup_experimental_pattern_tree_view(self, store, widget):
        """
            Creates the experimental data TreeView layout and behavior
        """
        setup_treeview(widget, store,
            on_cursor_changed=self.on_exp_data_tv_cursor_changed,
            sel_mode='MULTIPLE')
        store.connect('columns_changed', self.on_exp_columns_changed)
        self.update_exp_treeview(widget)
        # Other properties:
        self.exp_line_ctrl = LinePropertiesController(model=self.model.experimental_pattern, view=self.view.exp_line_view, parent=self)

    def setup_calculated_pattern_tree_view(self, store, widget):
        """
            Creates the calculated data TreeView layout and behavior
        """
        setup_treeview(widget, store,
            on_cursor_changed=self.on_exp_data_tv_cursor_changed,
            sel_mode='NONE')
        store.connect('columns_changed', self.on_calc_columns_changed)
        self.update_calc_treeview(widget)
        # Other properties:
        self.calc_line_ctrl = LinePropertiesController(model=self.model.calculated_pattern, view=self.view.calc_line_view, parent=self)

    def setup_exclusion_ranges_tree_view(self, store, widget):
        """
            Creates the exclusion ranges TreeView layout and behavior
        """
        setup_treeview(widget, store,
            on_cursor_changed=self.on_exclusion_ranges_tv_cursor_changed,
            sel_mode='MULTIPLE')
        
        def data_func(col, cell, model, iter, colnr):
            cell.set_property("text", "%g" % model.get(iter, colnr)[0])
        
        widget.append_column(new_text_column(
            'From [°2θ]', text_col=store.c_x, editable=True,
            data_func = (data_func, (store.c_x,)),
            edited_callback=(self.on_xy_data_cell_edited, (self.model.exclusion_ranges, 0)),
            resizable=True, expand=True))
        widget.append_column(new_text_column(
            'To [°2θ]', text_col=store.c_y, editable=True,
            data_func = (data_func, (store.c_y,)),
            edited_callback=(self.on_xy_data_cell_edited, (self.model.exclusion_ranges, 1)),
            resizable=True, expand=True))

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_experimental_pattern_tree_model(self):
        return wrap_xydata_to_treemodel(self.model, type(self.model).experimental_pattern)
    def get_calculated_pattern_tree_model(self):
        return wrap_xydata_to_treemodel(self.model, type(self.model).calculated_pattern)
    def get_exclusion_ranges_tree_model(self):
        return wrap_xydata_to_treemodel(self.model, type(self.model).exclusion_ranges)

    #used to keep a permanent fix on a child controller, prevents early GC
    _child_ctrl_ref = None

    def update_calc_treeview(self, tv):
        """
            Updates the calculated pattern TreeView layout
        """
        model = self.get_calculated_pattern_tree_model()

        for column in tv.get_columns():
            tv.remove_column(column)

        def get_num(column, cell, model, itr, col_id):
            cell.set_property('text', '%.3f' % model.get_value(itr, col_id))

        tv.append_column(new_text_column('2θ', data_func=(get_num, (model.c_x,))))
        tv.append_column(new_text_column('Cal', data_func=(get_num, (model.c_y,)) ))
        for i in range(model.get_n_columns() - 2):
            tv.append_column(new_text_column(
                self.model.calculated_pattern.get_y_name(i), data_func=(get_num, (i+2,))))

    def update_exp_treeview(self, tv):
        """
            Updates the experimental pattern TreeView layout
        """
        model = self.get_experimental_pattern_tree_model()

        for column in tv.get_columns():
            tv.remove_column(column)

        def get_num(column, cell, model, itr, col_id):
            cell.set_property('text', '%.3f' % model.get_value(itr, col_id))
        
        n_columns = model.get_n_columns()
        
        if n_columns > 2:
            for i in range(n_columns):
                tv.append_column(new_text_column(
                    self.model.calculated_pattern.get_y_name(i), editable=True,
                    edited_callback=(self.on_xy_data_cell_edited, (self.model.experimental_pattern, i)),
                    data_func=(get_num, (i,))
                ))
        else:
            # X Column:
            tv.append_column(new_text_column(
                '°2θ', editable=True,
                data_func=(get_num, (model.c_x,)),
                edited_callback=(self.on_xy_data_cell_edited, (self.model.experimental_pattern, 0))))
            # Y Column:
            tv.append_column(new_text_column(
                'Intensity', editable=True,
                data_func=(get_num, (model.c_y,)),
                edited_callback=(self.on_xy_data_cell_edited, (self.model.experimental_pattern, 1))))

    def remove_background(self):
        """
            Opens the 'remove background' dialog.
        """
        bg_view = BackgroundView(parent=self.view)
        self._child_ctrl_ref = BackgroundController(model=self.model.experimental_pattern, view=bg_view, parent=self)
        bg_view.present()

    def add_noise(self):
        """
            Opens the 'add noise' dialog.
        """
        an_view = AddNoiseView(parent=self.view)
        self._child_ctrl_ref = AddNoiseController(model=self.model.experimental_pattern, view=an_view, parent=self)
        an_view.present()

    def smooth_data(self):
        """
            Opens the 'smooth data' dialog.
        """
        sd_view = SmoothDataView(parent=self.view)
        self._child_ctrl_ref = SmoothDataController(model=self.model.experimental_pattern, view=sd_view, parent=self)
        sd_view.present()

    def shift_data(self):
        """
            Opens the 'shift data' dialog.
        """
        sh_view = ShiftDataView(parent=self.view)
        self._child_ctrl_ref = ShiftDataController(model=self.model.experimental_pattern, view=sh_view, parent=self)
        sh_view.present()

    def strip_peak(self):
        """
            Opens the 'strip peak' dialog.
        """
        st_view = StripPeakView(parent=self.view)
        self._child_ctrl_ref = StripPeakController(model=self.model.experimental_pattern, view=st_view, parent=self)
        st_view.present()

    def peak_properties(self):
        """
            Opens the 'peak properties' dialog.
        """
        pa_view = CalculatePeakPropertiesView(parent=self.view)
        self._child_ctrl_ref = CalculatePeakPropertiesController(model=self.model.experimental_pattern, view=pa_view, parent=self)
        pa_view.present()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_calc_columns_changed(self, *args, **kwargs):
        self.update_calc_treeview(self.view["specimen_calculated_pattern"])

    def on_exp_columns_changed(self, *args, **kwargs):
        self.update_exp_treeview(self.view["specimen_experimental_pattern"])

    def on_btn_ok_clicked(self, event):
        self.parent.pop_status_msg('edit_specimen')
        return super(SpecimenController, self).on_btn_ok_clicked(event)

    def on_exclusion_ranges_tv_cursor_changed(self, tv):
        path, col = tv.get_cursor()  # @UnusedVariable
        self.view["btn_del_exclusion_ranges"].set_sensitive(path is not None)
        return True

    def on_exp_data_tv_cursor_changed(self, tv):
        path, col = tv.get_cursor()  # @UnusedVariable
        self.view["btn_del_experimental_data"].set_sensitive(path is not None)
        return True

    def on_add_experimental_data_clicked(self, widget):
        self.model.experimental_pattern.append(0, 0)
        return True

    def on_add_exclusion_range_clicked(self, widget):
        self.model.exclusion_ranges.append(0, 0)
        return True

    def on_del_experimental_data_clicked(self, widget):
        paths = self.get_selected_paths(self.view["specimen_experimental_pattern"])
        if paths is not None:
            self.model.experimental_pattern.remove_from_indeces(*paths)
        return True

    def on_del_exclusion_ranges_clicked(self, widget):
        paths = self.get_selected_paths(self.view["specimen_exclusion_ranges"])
        if paths is not None:
            self.model.exclusion_ranges.remove_from_indeces(*paths)
        return True

    def on_xy_data_cell_edited(self, cell, path, new_text, model, col):
        try:
            value = float(locale.atof(new_text))
        except ValueError:
            logger.exception("ValueError: Invalid literal for float(): '%s'" % new_text)
        else:
            model.set_value(int(path), col, value)
        return True

    def on_import_exclusion_ranges_clicked(self, widget, data=None):
        def on_confirm(dialog):
            def on_accept(dialog):
                filename = dialog.filename
                parser = dialog.parser
                message = "An unexpected error has occured when trying to parse %s:\n\n<i>" % os.path.basename(filename)
                message += "{}</i>\n\n"
                message += "This is most likely caused by an invalid or unsupported file format."

                with DialogFactory.error_dialog_handler(message, parent=self.view.get_toplevel(), reraise=False):
                    self.model.exclusion_ranges.load_data(parser, filename, clear=True)
            DialogFactory.get_load_dialog(
                title="Import exclusion ranges", parent=self.view.get_top_widget(),
                filters=exc_parsers.get_import_file_filters()
            ).run(on_accept)
        DialogFactory.get_confirmation_dialog(
            "Importing exclusion ranges will erase all current data.\nAre you sure you want to continue?",
            parent=self.view.get_top_widget()
        ).run(on_confirm)

    def on_export_exclusion_ranges_clicked(self, widget, data=None):
        def on_accept(dialog):
            filename = dialog.filename
            parser = dialog.parser
            message = "An unexpected error has occurred when trying to save '%s'.\n" % os.path.basename(filename)
            message += "Contact the developer about this!"
            with DialogFactory.error_dialog_handler(message, parent=self.view.get_toplevel(), reraise=False):
                header = "%s %s" % (self.model.name, self.model.sample_name)
                self.model.exclusion_ranges.save_data(parser, filename, header=header)
        DialogFactory.get_save_dialog(
            "Select file for exclusion ranges export",
            parent=self.view.get_top_widget(), filters=exc_parsers.get_export_file_filters()
        ).run(on_accept)

    def on_replace_experimental_data(self, *args, **kwargs):
        def on_accept(dialog):
            filename = dialog.filename
            parser = dialog.parser
            message = "An unexpected error has occurred when trying to parse '%s'.\n" % os.path.basename(filename)
            message += "This is most likely caused by an invalid or unsupported file format."
            with DialogFactory.error_dialog_handler(message, parent=self.view.get_toplevel(), reraise=False):
                self.model.experimental_pattern.load_data(parser, filename, clear=True)
        DialogFactory.get_load_dialog(
            "Open XRD file for import", parent=self.view.get_top_widget(),
            filters=xrd_parsers.get_import_file_filters()
        ).run(on_accept)
        return True

    def on_btn_import_experimental_data_clicked(self, widget, data=None):
        def on_confirm(dialog):
            self.on_replace_experimental_data()
        DialogFactory.get_confirmation_dialog(
            "Importing a new experimental file will erase all current data.\nAre you sure you want to continue?",
            parent=self.view.get_top_widget()
        ).run(on_confirm)
        return True

    def on_export_experimental_data(self, *args, **kwargs):
        return self._export_data(self.model.experimental_pattern)

    def on_btn_export_experimental_data_clicked(self, widget, data=None):
        return self.on_export_experimental_data()

    def on_btn_export_calculated_data_clicked(self, widget, data=None):
        return self._export_data(self.model.calculated_pattern)

    def _export_data(self, line):
        def on_accept(dialog):
            filename = dialog.filename
            parser = dialog.parser
            message = "An unexpected error has occurred when trying to save to '%s'." % os.path.basename(filename)
            with DialogFactory.error_dialog_handler(message, parent=self.view.get_toplevel(), reraise=False):
                line.save_data(parser, filename, **self.model.get_export_meta_data())
        ext_less_fname = os.path.splitext(self.model.name)[0]
        DialogFactory.get_save_dialog(
            "Select file for export", parent=self.view.get_top_widget(),
            filters=xrd_parsers.get_export_file_filters(), current_name=ext_less_fname
        ).run(on_accept)

    pass # end of class

class StatisticsController(BaseController):

    def register_adapters(self):
        if self.model is not None:
            for name in self.model.get_properties():
                if name in self.model.__have_no_widget__:
                    pass
                else:
                    self.adapt(name)
            return

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from ...support.utils import not_none
from .basic import GtkAdapter


class ComboBoxAdapter(GtkAdapter):
    """
        An adapter that adapts a ComboBox widget to an property which has a 
        choices attribute containing dictionary with allowed (value, description)
        pairs as keys and values.
    """
    widget_types = ["option_list", ]
    _check_widget_type = Gtk.ComboBox

    _wid_read = lambda c, w, *a: Gtk.ComboBox.get_active_iter(w, *a)
    _wid_write = lambda c, w, *a: Gtk.ComboBox.set_active_iter(w, *a)
    _signal = "changed"

    _prop_cast = False

    def _parse_prop(self, prop, model):
        """Parses (optional) prop strings for the given model"""
        prop, model = super(ComboBoxAdapter, self)._parse_prop(prop, model)
        if not isinstance(prop.choices, dict):
            raise ValueError("ComboBox widget handler requires a property with a 'choices' dictionary!")
        else:
            self._store = Gtk.ListStore(str, str)
            for key, value in prop.choices.items():
                self._store.append([str(key), str(value)])
        return prop, model

    def _prop_write(self, itr):
        if itr is not None:
            return self._store.get_value(itr, 0)

    def _prop_read(self, val):
        for row in self._store:
            if self._store.get_value(row.iter, 0) == str(val):
                return row.iter

    def _connect_widget(self):
        # Set up the combo box layout:
        cell = Gtk.CellRendererText()
        self._widget.clear()
        self._widget.pack_start(cell, True)
        self._widget.add_attribute(cell, 'text', 1)
        cell.set_property('family', 'Monospace')
        cell.set_property('size-points', 10)

        # Set the model:
        self._widget.set_model(self._store)

        # Continue as usual:
        super(ComboBoxAdapter, self)._connect_widget()

    def disconnect(self, model=None, widget=None):
        widget = not_none(self._widget, widget)
        if widget is not None: widget.set_model(None)
        super(ComboBoxAdapter, self).disconnect(model=model, widget=widget)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2016 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import os

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from pyxrd.generic.io.utils import get_case_insensitive_glob

def get_file_chooser_kwags():
    return dict(
        action=Gtk.FileChooserAction.SAVE,
        title="The dialog title",
        parent=Gtk.Window(),
        current_name="suggested_file_name",
        current_folder=os.path.expanduser("~"),
        extra_widget=Gtk.Label(label="Test Label"),
        filters=[ ("Text File", get_case_insensitive_glob("*.txt")) ],
        multiple=False,
        confirm_overwrite=True
    )



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Author: Mathijs Dumon
# ThreadedTaskBox based on code from Rick Spencer
# This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send
# a letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

import gi   
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from mvc.support.cancellable_thread import CancellableThread
from mvc.support.gui_loop import add_timeout_call, remove_timeout_call,\
    run_when_idle


class ThreadedTaskBox(Gtk.Table):
    """
        ThreadedTaskBox: encapsulates a spinner, label and a cancel button for
        threaded tasks.
    """

    __gsignals__ = {
        'cancelrequested' : (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)), #@UndefinedVariable
        'stoprequested' : (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)) #@UndefinedVariable
    }

    def __init__(self, cancelable=True, stoppable=False):
        """
            Create a ThreadedTaskBox

            Keyword arguments:
            cancelable -- optional value to determine whether to show cancel button. Defaults to True.
            stoppable -- optional value to determine whether to show the stop button. Default to False.
        """
        super(ThreadedTaskBox, self).__init__()
        self.setup_ui(cancelable=cancelable, stoppable=stoppable)

    def setup_ui(self, cancelable=True, stoppable=False):
        GObject.GObject.__init__(self, 3, 3)
        self.set_row_spacings(10)
        self.set_col_spacings(10)

        self.descrlbl = Gtk.Label(label="Status:")
        self.descrlbl.show()
        self.attach(self.descrlbl, 0, 3, 0, 1, xoptions=Gtk.AttachOptions.FILL, yoptions=0)

        self.spinner = Gtk.Spinner()
        self.spinner.show()
        self.attach(self.spinner, 0, 1, 1, 3, xoptions=0, yoptions=0)

        self.label = Gtk.Label()
        self.label.show()
        self.attach(self.label, 1, 2, 1, 3, xoptions=Gtk.AttachOptions.FILL, yoptions=0)

        self.cancel_button = Gtk.Button(stock=Gtk.STOCK_CANCEL)
        self.cancel_button.set_sensitive(False)
        self.cancel_button.connect("clicked", self.__cancel_clicked)
        if cancelable:
            self.attach(self.cancel_button, 2, 3, 1, 2, xoptions=0, yoptions=0)

        self.stop_button = Gtk.Button(stock=Gtk.STOCK_STOP)
        self.stop_button.set_sensitive(False)
        self.stop_button.connect("clicked", self.__stop_clicked)
        if stoppable:
            self.attach(self.stop_button, 2, 3, 2, 3, xoptions=0, yoptions=0)

        self.set_no_show_all(False)
        self.set_visible(True)
        self.show_all()

    def start(self):
        # Start the spinner
        self.spinner.start()

        # Enable the buttons so the user can try to cancel the task
        self.cancel_button.set_sensitive(True)
        self.stop_button.set_sensitive(True)

    def set_status(self, caption):
        self.label.set_text(caption)

    def stop(self, join=False, cancel=False):
        """
            Stops spinning the spinner and emits the correct event.
        """
        # disable the cancel button since the task is about to be told to stop
        self.cancel_button.set_sensitive(False)
        self.stop_button.set_sensitive(False)

        if cancel:
            self.emit("cancelrequested", self)
        else:
            self.emit("stoprequested", self)

        self.spinner.stop()
        self.label.set_text("Done")

    def cancel(self):
        self.stop(cancel=True)

    def __cancel_clicked(self, widget):
        self.cancel()

    def __stop_clicked(self, widget):
        self.stop()

    pass # end of class

GObject.type_register(ThreadedTaskBox) #@UndefinedVariable



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from contextlib import contextmanager

from mvc.models.base import Model

from pyxrd.generic.controllers import ObjectListStoreController

from pyxrd.mixture.models import Mixture #, InSituMixture
from pyxrd.mixture.views import EditMixtureView, AddMixtureView #EditInSituMixtureView,

from .edit_mixture_controller import EditMixtureController
#from .edit_insitu_mixture_controller import EditInSituMixtureController
from .add_mixture_controller import AddMixtureController


class MixturesController(ObjectListStoreController):

    treemodel_property_name = "mixtures"
    treemodel_class_type = Mixture
    columns = [ ("Mixture name", "c_name") ]
    delete_msg = "Deleting a mixture is irreverisble!\nAre You sure you want to continue?"
    obj_type_map = [
        #(InSituMixture, EditInSituMixtureView, EditInSituMixtureController),
        (Mixture, EditMixtureView, EditMixtureController),
    ]

    def get_mixtures_tree_model(self, *args):
        return self.treemodel

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_load_object_clicked(self, event):
        pass # cannot load mixtures
    
    def on_save_object_clicked(self, event):
        pass # cannot save mixtures

    def create_new_object_proxy(self):
        """def on_accept(mixture_type):
            if mixture_type == "mixture":
                self.add_object(Mixture(parent=self.model))
            #elif mixture_type == "insitu":
            #    self.add_object(InSituMixture(parent=self.model))
                
        # TODO re-use this and reset the COMBO etc.
        self.add_model = Model()
        self.add_view = AddMixtureView(types_dict={
            'mixture': "Create a regular mixture", 
            #'insitu': "Create an in-situ mixture"
        }, parent=self.view)
        self.add_ctrl = AddMixtureController(
            model=self.add_model, view=self.add_view, parent=self.parent,
            callback=on_accept
        )

        self.add_view.present()"""
        return Mixture(parent=self.model) #None

    @contextmanager
    def _multi_operation_context(self):
        with self.model.data_changed.hold():
            yield

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from io import *
from .custom_io import *
from .utils import get_case_insensitive_glob, retrieve_lowercase_extension



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from weakref import WeakValueDictionary
import threading
import multiprocessing
import logging
logger = logging.getLogger(__name__)

from ..support.utils import get_new_uuid

class ObjectPool(object):
    """
        This class allows to fetch mvc model objects using their UUID.
        This requires to model to have a property called "uuid". All
        class inheriting from the base 'Model' class will have this.
        If implementing a custom model, the UUID property is responsible
        for the removal and addition to the pool when it changes values.
        Also see the UUIDProperty descriptor for an example implementation.
        We can use this to store complex relations between objects where 
        references to each other can be replaced with the UUID.
        For a multi-threaded version see ThreadedObjectPool. 
    """

    def __init__(self, *args, **kwargs):
        object.__init__(self)
        self._objects = WeakValueDictionary()

    def add_or_get_object(self, obj):
        try:
            self.add_object(obj, force=False, silent=False)
            return obj
        except KeyError:
            return self.get_object(obj.uuid)

    def add_object(self, obj, force=False, fail_on_duplicate=False):
        if not obj.uuid in self._objects or force:
            self._objects[obj.uuid] = obj
        elif fail_on_duplicate:
            raise KeyError("UUID %s is already taken by another object %s, cannot add object %s" % (obj.uuid, self._objects[obj.uuid], obj))
        else:
            # Just change the objects uuid, will break refs, but
            # it prevents issues with inherited properties etc.
            logger.warning("A duplicate UUID was passed to an ObjectPool for a %s object." % obj)
            obj.uuid = get_new_uuid()

    def change_all_uuids(self):
        # first get a copy off all uuids & objects:
        items = list(self._objects.items())
        for uuid, obj in items: # @UnusedVariable
            obj.uuid = get_new_uuid()

    def remove_object(self, obj):
        if obj.uuid in self._objects and self._objects[obj.uuid] == obj:
            del self._objects[obj.uuid]

    def get_object(self, uuid):
        obj = self._objects.get(uuid, None)
        return obj

    def clear(self):
        self._objects.clear()

class ThreadedObjectPool(object):

    def __init__(self, *args, **kwargs):
        object.__init__(self)
        self.pools = {}

    def clean_pools(self):
        for ptkey in list(self.pools.keys()):
            if (ptkey == (None, None) or not ptkey[0].is_alive() or not ptkey[1].is_alive()):
                del self.pools[ptkey] # clear this sucker

    def get_pool(self):
        process = multiprocessing.current_process()
        thread = threading.current_thread()
        pool = self.pools.get((process, thread), ObjectPool())
        self.pools[(process, thread)] = pool
        return pool

    def add_or_get_object(self, *args, **kwargs):
        pool = self.get_pool()
        return pool.add_or_get_object(*args, **kwargs)

    def add_object(self, *args, **kwargs):
        pool = self.get_pool()
        return pool.add_object(*args, **kwargs)

    def change_all_uuids(self, *args, **kwargs):
        pool = self.get_pool()
        return pool.change_all_uuids(*args, **kwargs)

    def remove_object(self, *args, **kwargs):
        pool = self.get_pool()
        return pool.remove_object(*args, **kwargs)

    def get_object(self, *args, **kwargs):
        pool = self.get_pool()
        return pool.get_object(*args, **kwargs)

    def clear(self, *args, **kwargs):
        pool = self.get_pool()
        return pool.clear(*args, **kwargs)

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon

from .dummy_async_server import DummyAsyncServer

class DummyAsyncServerProvider(object):
    
    _server = DummyAsyncServer()
    
    @classmethod
    def get_status(cls):
        """ should return a three-tuple consisting of the status colour, label and a description:
            ("#FF0000", "Error", "Nameserver not running")
        """            
        return ("#00FF00", "Connected (Dummy)", "Succesfully connected to Dummy PyXRD Server")
    
    @classmethod
    def get_server(cls):
        return cls._server
    
    @classmethod
    def launch_server(cls):
        if cls._server is None:
            cls._server = DummyAsyncServer()
        
    @classmethod
    def stop_server(cls):
        cls._server.shutdown()
        del cls._server
        
    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import re
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from .entry_adapter import EntryAdapter

class FloatEntryAdapter(EntryAdapter):
    """
        An adapter for a Gtk.Entry widget holding a float.
    """
    widget_types = ["float_entry", "float_input"]
    _check_widget_type = Gtk.Entry
    _signal = "changed"

    def __init__(self, *args, **kwargs):
        super(FloatEntryAdapter, self).__init__(*args, **kwargs)
        numeric_const_pattern = r"""
        [-+]? # optional sign
        (?:
            (?: \d* \. \d+ ) # .1 .12 .123 etc 9.1 etc 98.1 etc
            |
            (?: \d+ \.? ) # 1. 12. 123. etc 1 12 123 etc
        )
        # followed by optional exponent part if desired
        (?: [Ee] [+-]? \d+ ) ?
        """
        self.rx = re.compile(numeric_const_pattern, re.VERBOSE)

    def _prop_read(self, *args):
        return str(*args)

    def _prop_write(self, *args):
        try:
            return float(*args)
        except ValueError:
            return self._get_property_value()

    def _on_wid_changed(self, widget, *args):
        """Called when the widget is changed"""
        with self._block_widget_signal():
            if self._ignoring_notifs: return
            self._validate_float(widget)
            super(FloatEntryAdapter, self)._on_wid_changed(widget, *args)

    def _validate_float(self, entry):
        entry_text = entry.get_text()
        newtext = self.rx.findall(entry_text)
        if len(newtext) > 0:
            entry.set_text(newtext[0])
        else:
            entry.set_text("")

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.generic.controllers import BaseController, DialogController
from pyxrd.generic.plot.eye_dropper import EyeDropper
from pyxrd.file_parsers.xrd_parsers import xrd_parsers
from pyxrd.data import settings

class LinePropertiesController(BaseController):
    """
        Controller for Line models' general properties
    """
    auto_adapt_excluded = [
        "noise_fraction",
        "shift_value",
        "shift_position",
        "smooth_type",
        "smooth_degree",
        "strip_startx",
        "strip_endx",
        "noise_level",
        "bg_type",
        "bg_position",
        "bg_scale",
        "peak_startx",
        "peak_endx",
        "peak_area_result",
        "peak_fwhm_result"
    ] # these are handled by other controllers

    __LINEID__ = "TRUE"

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def register_adapters(self):
        super(LinePropertiesController, self).register_adapters()
        self.update_sensitivities()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update_sensitivities(self):
        """
            Updates the views sensitivities according to the model state.
        """
        self.view[self.view.widget_format % "color"].set_sensitive(not self.model.inherit_color)
        self.view["spb_%s" % self.view.widget_format % "lw"].set_sensitive(not self.model.inherit_lw)
        self.view[self.view.widget_format % "ls"].set_sensitive(not self.model.inherit_ls)
        self.view[self.view.widget_format % "marker"].set_sensitive(not self.model.inherit_marker)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @BaseController.observe("inherit_color", assign=True)
    @BaseController.observe("inherit_lw", assign=True)
    @BaseController.observe("inherit_ls", assign=True)
    @BaseController.observe("inherit_marker", assign=True)
    def notif_color_toggled(self, model, prop_name, info):
        self.update_sensitivities()

    pass  # end of class


class PatternActionController(DialogController):
    """
        General class for actions that can be applied on ExperimentalPattern
        models.
        
        Attributes:
         model_setup_method: the name of the method to call on the model to
          auto-setup any variables for the action. Optional; can be None.
        model_action_method: the name of the method to call on the model when
         the user decides to apply the action. Required, cannot be None.
        model_cancel_method: the name of the method to call on the model when
         the user decides to cancel the action. Optional; can be None.
    """
    model_setup_method = None
    model_action_method = None
    model_cancel_method = None

    def register_adapters(self):
        if self.model_setup_method is not None:
            getattr(self.model, self.model_setup_method)()
        return super(PatternActionController, self).register_adapters()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_btn_ok_clicked(self, event):
        try:
            action = getattr(self.model, self.model_action_method)
        except AttributeError:
            raise ValueError("Subclasses of PatternActionController should specify a valid action method name (was '%s')!" % self.model_action_method)
        else:
            action()
        return super(PatternActionController, self).on_btn_ok_clicked(event)

    def on_cancel(self):
        if self.model_cancel_method is not None:
            getattr(self.model, self.model_cancel_method)()
        return super(PatternActionController, self).on_cancel()

    pass # end of class

class AddNoiseController(PatternActionController):
    """
        Controller for the experimental pattern 'add noise' action view.
    """

    auto_adapt_included = [
        "noise_fraction",
    ]
    model_setup_method = None
    model_action_method = "add_noise"
    model_cancel_method = "clear_noise_variables"

    pass # end of class


class ShiftDataController(PatternActionController):
    """
        Controller for the experimental pattern 'shift data' action view.
    """
    auto_adapt_included = [
        "shift_value",
        "shift_position",
    ]
    model_setup_method = "setup_shift_variables"
    model_action_method = "shift_data"
    model_cancel_method = "clear_shift_variables"

    pass # end of class

class SmoothDataController(PatternActionController):

    auto_adapt_included = [
        "smooth_type",
        "smooth_degree",
    ]
    model_setup_method = "setup_smooth_variables"
    model_action_method = "smooth_data"
    model_cancel_method = "clear_smooth_variables"

    pass # end of class

class CalculatePeakPropertiesController(PatternActionController):

    auto_adapt_included = [
        "peak_startx",
        "peak_endx",
        "peak_fwhm_result",
        "peak_area_result"
    ]
    model_setup_method = None
    model_action_method = "clear_peak_properties_variables"
    model_cancel_method = "clear_peak_properties_variables"

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_sample_start_clicked(self, event):
        self.sample("peak_startx")
        return True

    def on_sample_end_clicked(self, event):
        self.sample("peak_endx")
        return True

    def sample(self, attribute):

        def onclick(x_pos, *args):
            if self.edc is not None:
                self.edc.enabled = False
                self.edc.disconnect()
                self.edc = None
            if x_pos != -1:
                setattr(self.model, attribute, x_pos)
            self.view.get_toplevel().present()

        self.edc = EyeDropper(
            self.parent.parent.plot_controller,
            onclick
        )

        self.view.get_toplevel().hide()
        self.parent.parent.view.get_toplevel().present()

    pass # end of class

class StripPeakController(PatternActionController):

    auto_adapt_included = [
        "strip_startx",
        "strip_endx",
        "noise_level",
    ]
    model_setup_method = None
    model_action_method = "strip_peak"
    model_cancel_method = "clear_strip_variables"

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_sample_start_clicked(self, event):
        self.sample("strip_startx")
        return True

    def on_sample_end_clicked(self, event):
        self.sample("strip_endx")
        return True

    def sample(self, attribute):

        def onclick(x_pos, *args):
            if self.edc is not None:
                self.edc.enabled = False
                self.edc.disconnect()
                self.edc = None
            if x_pos != -1:
                setattr(self.model, attribute, x_pos)
            self.view.get_toplevel().present()

        self.edc = EyeDropper(
            self.parent.parent.plot_controller,
            onclick
        )

        self.view.get_toplevel().hide()
        self.parent.parent.view.get_toplevel().present()

    pass # end of class

class BackgroundController(PatternActionController):
    """
        Controller for the experimental pattern 'remove background' action view.
    """

    file_filters = xrd_parsers.get_import_file_filters()

    auto_adapt_included = [
        "bg_type",
        "bg_position",
        "bg_scale",
    ]
    model_setup_method = "find_bg_position"
    model_action_method = "remove_background"
    model_cancel_method = "clear_bg_variables"

    def register_view(self, view):
        super(BackgroundController, self).register_view(view)
        view.set_file_dialog(
            DialogFactory.get_load_dialog(
                title="Open XRD file for import",
                parent=view.get_top_widget(),
                filters=self.file_filters
            ),
            self.on_pattern_file_set
        )
        view.select_bg_view(self.model.get_bg_type_lbl().lower())

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @PatternActionController.observe("bg_type", assign=True)
    def notif_bg_type_changed(self, model, prop_name, info):
        self.view.select_bg_view(self.model.get_bg_type_lbl().lower())
        return

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_pattern_file_set(self, button, dialog):
        # TODO
        # This should allow more flexibility:
        #  Patterns should be allowed to not have the exact same shape,
        #  add an x-shift variable to align them
        filename = dialog.filename
        parser = dialog.parser
        data_objects = None

        message = "An unexpected error has occurred when trying to parse %s:\n\n<i>" % os.path.basename(filename)
        message += "{}</i>\n\n"
        message += "This is most likely caused by an invalid or unsupported file format."
        with DialogFactory.error_dialog_handler(message, parent=self.view.get_toplevel(), reraise=settings.DEBUG):
            # Parse the pattern file
            data_objects = parser.parse(filename)
            pattern = data_objects[0].data
            bg_pattern_x = pattern[:, 0].copy()
            bg_pattern_y = pattern[:, 1].copy()
            del pattern

            # Interpolate/Extrapolate where needed to match data shape and range
            from scipy.interpolate import interp1d
            f = interp1d(
                bg_pattern_x, bg_pattern_y,
                bounds_error=False, fill_value=0
            )
            bg_xnew = self.model.data_x
            bg_ynew = f(bg_xnew)
            self.model.bg_pattern = bg_ynew

    pass # end of class
